from abc import ABC, abstractmethod
from typing import Type, TypeVar, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from dlt.pipeline.trace import PipelineTrace, PipelineStepTrace
    from dlt.pipeline.typing import TPipelineStep
    from dlt.common.pipeline import SupportsPipeline
else:
    PipelineTrace = PipelineStepTrace = TPipelineStep = SupportsPipeline = Any

from dlt.common.runtime.tracking import SupportsTracking


TCollector = TypeVar("TCollector", bound="Collector")


class Collector(ABC, SupportsTracking):
    step: str

    @abstractmethod
    def update(
        self,
        name: str,
        inc: int = 1,
        total: int = None,
        inc_total: int = None,
        message: str = None,
        label: str = None,
    ) -> None:
        """Creates or updates a counter

        This function updates a counter `name` with a value `inc`. If counter does not exist, it is
        created with optional total value of `total`. Labels are used to create sub-counters.

        Examples for counters by stage:
        ```python
        # Extract Stage
        collector.update("users", inc=5)           # 5 rows added to users table
        collector.update("Resources", inc=1)       # 1 resource processed
        collector.update("Resources", inc=1, label="Completed")  # 1 resource completed

        # Normalize Stage
        collector.update("Files", inc=1, total=10)     # 1/10 files processed
        collector.update("Items", inc=100)             # 100 items processed

        # Load Stage
        collector.update("Jobs", inc=1, total=5)       # 1/5 jobs processed
        collector.update("Jobs", inc=1, label="Failed")  # 1 job failed
        ```
        """
        pass

    @abstractmethod
    def _start(self, step: str) -> None:
        """Starts counting for a processing step with name `step`"""
        pass

    @abstractmethod
    def _stop(self) -> None:
        """Stops counting. Should close all counters and release resources ie. screen or push the results to a server."""
        pass

    def __call__(self: TCollector, step: str) -> TCollector:
        """Syntactic sugar for nicer context managers"""
        self.step = step
        return self

    def __enter__(self: TCollector) -> TCollector:
        self._start(self.step)
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: Any) -> None:
        self._stop()

    def on_start_trace(
        self, trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
    ) -> None:
        pass

    def on_start_trace_step(
        self, trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
    ) -> None:
        pass

    def on_end_trace_step(
        self,
        trace: PipelineTrace,
        step: PipelineStepTrace,
        pipeline: SupportsPipeline,
        step_info: Any,
        send_state: bool,
    ) -> None:
        pass

    def on_end_trace(
        self, trace: PipelineTrace, pipeline: SupportsPipeline, send_state: bool
    ) -> None:
        pass



from typing import Any, Dict, Type, Union, TYPE_CHECKING, Optional

from dlt.common.data_types.typing import TDataType
from dlt.common.destination import Destination, DestinationCapabilitiesContext, PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.data_writers.escape import escape_postgres_identifier, escape_mssql_literal
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.schema.typing import TColumnSchema
from dlt.common.typing import TLoaderFileFormat

from dlt.destinations.impl.mssql.factory import MsSqlTypeMapper
from dlt.destinations.impl.synapse.configuration import (
    SynapseCredentials,
    SynapseClientConfiguration,
)
from dlt.destinations.impl.synapse.synapse_adapter import TTableIndexType

if TYPE_CHECKING:
    from dlt.destinations.impl.synapse.synapse import SynapseClient


class SynapseTypeMapper(MsSqlTypeMapper):
    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        # TIME is not supported for parquet
        if loader_file_format == "parquet" and column["data_type"] == "time":
            raise TerminalValueError(
                "Please convert `datetime.time` objects in your data to `str` or"
                " `datetime.datetime`.",
                "time",
            )

    def to_destination_type(self, column: TColumnSchema, table: PreparedTableSchema) -> str:
        sc_t = column["data_type"]
        if sc_t == "json":
            return "nvarchar(%s)" % column.get("precision", "max")
        return super().to_destination_type(column, table)


class synapse(Destination[SynapseClientConfiguration, "SynapseClient"]):
    spec = SynapseClientConfiguration

    # TODO: implement as property everywhere and makes sure not accessed as class property
    # @property
    # def spec(self) ->  Type[SynapseClientConfiguration]:
    #     return SynapseClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()

        caps.preferred_loader_file_format = "insert_values"
        caps.supported_loader_file_formats = ["insert_values", "model"]
        caps.preferred_staging_file_format = "parquet"
        caps.supported_staging_file_formats = ["parquet"]
        caps.type_mapper = SynapseTypeMapper

        caps.insert_values_writer_type = "select_union"  # https://stackoverflow.com/a/77014299

        # similarly to mssql case sensitivity depends on database collation
        # https://learn.microsoft.com/en-us/sql/relational-databases/collations/collation-and-unicode-support?view=sql-server-ver16#collations-in-azure-sql-database
        # note that special option CATALOG_COLLATION is used to change it
        caps.escape_identifier = escape_postgres_identifier
        caps.escape_literal = escape_mssql_literal
        # we allow to reconfigure capabilities in the mssql factory
        caps.has_case_sensitive_identifiers = False

        # Synapse has a max precision of 38
        # https://learn.microsoft.com/en-us/sql/t-sql/statements/create-table-azure-sql-data-warehouse?view=aps-pdw-2016-au7#DataTypes
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)

        # https://learn.microsoft.com/en-us/sql/t-sql/statements/create-table-azure-sql-data-warehouse?view=aps-pdw-2016-au7#LimitationsRestrictions
        caps.max_identifier_length = 128
        caps.max_column_identifier_length = 128

        # https://learn.microsoft.com/en-us/azure/synapse-analytics/sql-data-warehouse/sql-data-warehouse-service-capacity-limits#queries
        caps.max_query_length = 65536 * 4096
        caps.is_max_query_length_in_bytes = True

        # nvarchar(max) can store 2 GB
        # https://learn.microsoft.com/en-us/sql/t-sql/data-types/nchar-and-nvarchar-transact-sql?view=sql-server-ver16#nvarchar---n--max--
        caps.max_text_data_type_length = 2 * 1024 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True

        # https://learn.microsoft.com/en-us/azure/synapse-analytics/sql-data-warehouse/sql-data-warehouse-develop-transactions
        caps.supports_transactions = True
        caps.supports_ddl_transactions = False

        caps.supports_create_table_if_not_exists = (
            False  # IF NOT EXISTS on CREATE TABLE not supported
        )
        caps.supports_multiple_statements = True

        # Synapse throws "Some part of your SQL statement is nested too deeply. Rewrite the query or break it up into smaller queries."
        # if number of records exceeds a certain number. Which exact number that is seems not deterministic:
        # in tests, I've seen a query with 12230 records run successfully on one run, but fail on a subsequent run, while the query remained exactly the same.
        # 10.000 records is a "safe" amount that always seems to work.
        caps.max_rows_per_insert = 10000

        # NOTE: datetimeoffset can store 7 digits for fractional seconds, maybe you could use it with parquet in ns
        #   precision. you can pass synapse(timestamp_precision=7) to override
        # https://learn.microsoft.com/en-us/sql/t-sql/data-types/datetimeoffset-transact-sql?view=sql-server-ver16
        caps.timestamp_precision = 6

        caps.supported_merge_strategies = ["delete-insert", "scd2"]
        caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]

        caps.sqlglot_dialect = "tsql"

        return caps

    @property
    def client_class(self) -> Type["SynapseClient"]:
        from dlt.destinations.impl.synapse.synapse import SynapseClient

        return SynapseClient

    def __init__(
        self,
        credentials: Union[SynapseCredentials, Dict[str, Any], str] = None,
        default_table_index_type: TTableIndexType = "heap",
        create_indexes: bool = False,
        staging_use_msi: bool = False,
        has_case_sensitive_identifiers: bool = False,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Synapse destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[SynapseCredentials, Dict[str, Any], str], optional): Credentials to connect to the Synapse dedicated pool. Can be an instance of `SynapseCredentials` or
                a connection string in the format `synapse://user:password@host:port/database`
            default_table_index_type (TTableIndexType): Maps directly to the default_table_index_type attribute of the SynapseClientConfiguration object.
            create_indexes (bool, optional): Maps directly to the create_indexes attribute of the SynapseClientConfiguration object.
            staging_use_msi (bool, optional): Maps directly to the staging_use_msi attribute of the SynapseClientConfiguration object.
            has_case_sensitive_identifiers (bool, optional): Are identifiers used by synapse database case sensitive (following the catalog collation)
            destination_name (str, optional): Name of the destination. Defaults to None.
            environment (str, optional): Environment name. Defaults to None.
            **kwargs (Any, optional): Additional arguments passed to the destination config
        """
        super().__init__(
            credentials=credentials,
            default_table_index_type=default_table_index_type,
            create_indexes=create_indexes,
            staging_use_msi=staging_use_msi,
            has_case_sensitive_identifiers=has_case_sensitive_identifiers,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: SynapseClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        # modify the caps if case sensitive identifiers are requested
        if config.has_case_sensitive_identifiers:
            caps.has_case_sensitive_identifiers = True
            caps.casefold_identifier = str
        return super().adjust_capabilities(caps, config, naming)


synapse.register()



from __future__ import annotations

from types import TracebackType
from typing import Any, Optional, Type, Union, TYPE_CHECKING, Literal, overload, Collection

from sqlglot.schema import Schema as SQLGlotSchema
import sqlglot.expressions as sge

import dlt
from dlt.common.destination.exceptions import OpenTableClientNotAvailable
from dlt.common.libs.sqlglot import TSqlGlotDialect
from dlt.common.json import json
from dlt.common.versioned_state import decompress_state
from dlt.common.destination.reference import AnyDestination, TDestinationReferenceArg, Destination
from dlt.common.destination.client import JobClientBase, SupportsOpenTables, WithStateSync
from dlt.common.schema import Schema
from dlt.common.typing import Self
from dlt.common.schema.typing import C_DLT_LOAD_ID, C_DLT_LOADS_TABLE_LOAD_ID, LOADS_TABLE_NAME
from dlt.common.utils import simple_repr, without_none
from dlt.common.exceptions import ValueErrorWithKnownValues
from dlt.destinations.sql_client import SqlClientBase, WithSqlClient
from dlt.dataset import lineage
from dlt.dataset.utils import get_destination_clients
from dlt.destinations.queries import build_row_counts_expr
from dlt.common.destination.exceptions import (
    DestinationUndefinedEntity,
    SqlClientNotAvailable,
)

if TYPE_CHECKING:
    from dlt.common.libs.ibis import ir
    from dlt.common.libs.ibis import BaseBackend as IbisBackend


class Dataset:
    """Access to dataframes and arrow tables in the destination dataset via dbapi"""

    def __init__(
        self,
        destination: TDestinationReferenceArg,
        dataset_name: str,
        schema: Union[dlt.Schema, str, None] = None,
    ) -> None:
        self._destination_reference = destination
        self._destination: AnyDestination = Destination.from_reference(destination)
        self._dataset_name = dataset_name
        self._schema: Union[dlt.Schema, str, None] = schema
        self._pipeline_name: Optional[str] = None
        """If _schema not provided, used to get default schema from state"""
        # self._sqlglot_schema: SQLGlotSchema = None
        self._sql_client: SqlClientBase[Any] = None
        self._opened_sql_client: SqlClientBase[Any] = None
        self._table_client: SupportsOpenTables = None

    def ibis(self, read_only: bool = False) -> IbisBackend:
        """Get an ibis backend for the dataset.

        This creates a connection to the destination.

        The `read_only` flag is currently only supported for duckdb destination.
        """
        from dlt.helpers.ibis import create_ibis_backend

        return create_ibis_backend(
            destination=self._destination,
            client=self.destination_client,
            read_only=read_only,
        )

    @property
    def schema(self) -> dlt.Schema:
        """dlt schema associated with the dataset.

        If no provided at dataset initialization, it is fetched from the destination. Fallbacks
        to local dlt pipeline metadata.
        """
        if isinstance(self._schema, dlt.Schema):
            return self._schema

        maybe_schema: Optional[dlt.Schema] = None

        if isinstance(self._schema, str):
            maybe_schema = _get_dataset_schema_from_destination_using_schema_name(
                self, self._schema
            )

        if not maybe_schema:
            maybe_schema = _get_dataset_schema_from_dataset_dlt_tables(self)

        if not maybe_schema:
            # uses local dlt pipeline data instead of destination
            maybe_schema = dlt.Schema(self.dataset_name)

        assert isinstance(maybe_schema, dlt.Schema)
        self._schema = maybe_schema
        return self._schema

    @property
    def tables(self) -> list[str]:
        """List of table names found in the dataset.

        This only includes "completed tables". In other words, during the lifetime of a `pipeline.run()`
        execution, tables may exist on the destination, but will only appear on the dataset once
        `pipeline.run()` is done.
        """
        # return only completed tables
        return self.schema.data_table_names() + self.schema.dlt_table_names()

    def _ipython_key_completions_(self) -> list[str]:
        """Provide table names as completion suggestion in interactive environments."""
        return self.tables

    @property
    def sqlglot_schema(self) -> SQLGlotSchema:
        """SQLGlot schema of the dataset derived from the dlt schema."""
        # NOTE: no cache for now, it is probably more expensive to compute the current schema hash
        # to see wether this is stale than to compute a new sqlglot schema
        return lineage.create_sqlglot_schema(
            self.schema, self.dataset_name, dialect=self.destination_dialect
        )

    @property
    def destination_dialect(self) -> TSqlGlotDialect:
        """SQLGlot dialect of the dataset destination.

        This is the target dialect when transpiling SQL queries.
        """
        return self.sql_client.capabilities.sqlglot_dialect

    @property
    def dataset_name(self) -> str:
        """Name of the dataset"""
        return self._dataset_name

    # TODO why do we need `_opened_sql_client` and `_sql_client`? One seems used by
    # the `dlt.Dataset` context manager and the other by `dlt.Relation`
    @property
    def sql_client(self) -> SqlClientBase[Any]:
        # return the opened sql client if it exists
        if self._opened_sql_client:
            return self._opened_sql_client
        if not self._sql_client:
            self._sql_client = get_dataset_sql_client(self)
        return self._sql_client

    # TODO remove this; this seems only implemented to pass to `Relation` for it
    # to respect the ABC requirements of `WithSqlClient` mixin
    # also, it's odd that we need to instantiate an SQL client to get its class
    @property
    def sql_client_class(self) -> Type[SqlClientBase[Any]]:
        return self.sql_client.__class__

    # TODO if `destination_client` returns a new client each time, it shouldn't be a property
    @property
    def destination_client(self) -> JobClientBase:
        return get_dataset_destination_client(self)

    # TODO should this public? This should only be accessed via `.__open__()`
    @property
    def open_table_client(self) -> SupportsOpenTables:
        if not self._table_client:
            client = get_dataset_destination_client(self)
            if isinstance(client, SupportsOpenTables):
                self._table_client = client
            else:
                raise OpenTableClientNotAvailable(
                    self.dataset_name, self._destination.destination_name
                )
        return self._table_client

    # TODO remove method; need to update `dlthub` to avoid conflict
    # this is only used by `dlt.hub.transformation` currently
    def is_same_physical_destination(self, other: dlt.Dataset) -> bool:
        """
        Returns true if the other dataset is on the same physical destination
        helpful if we want to run sql queries without extracting the data
        """
        return is_same_physical_destination(self, other)

    def query(
        self,
        query: Union[str, sge.Select, ir.Expr],
        query_dialect: Optional[TSqlGlotDialect] = None,
        *,
        _execute_raw_query: bool = False,
    ) -> dlt.Relation:
        """Create a `dlt.Relation` from an SQL query, SQLGlot expression or Ibis expression.

        Args:
            query (Union[str, sge.Select, ir.Expr]): The query that defines the relation.
            query_dialect (Optional[TSqlGlotDialect]): The dialect of the query. If specified, it will be used to transpile the query
                to the destination's dialect. Otherwise, the query is assumed to be the destination's dialect (accessible via `Dataset.sqlglot_dialect`)

        Returns:
            dlt.Relation: The relation for the query
        """
        return dlt.Relation(
            dataset=self,
            query=query,
            query_dialect=query_dialect,
            _execute_raw_query=_execute_raw_query,
        )

    # NOTE could simply accept `*args, **kwargs` and pass to `.query()` but would decrease readability
    #
    def __call__(
        self,
        query: Union[str, sge.Select, ir.Expr],
        query_dialect: Optional[TSqlGlotDialect] = None,
        *,
        _execute_raw_query: bool = False,
    ) -> dlt.Relation:
        """Convenience method to proxy `Dataset.query()`. See this method for details."""
        return self.query(query, query_dialect, _execute_raw_query=_execute_raw_query)

    def table(
        self, table_name: str, *, load_ids: Optional[Collection[str]] = None, **kwargs: Any
    ) -> dlt.Relation:
        """Get a `dlt.Relation` associated with a table from the dataset."""
        if table_name not in self.tables:
            # TODO: raise TableNotFound
            raise ValueError(f"Table `{table_name}` not found. Available table(s): {self.tables}")

        # TODO remove in due time;
        if kwargs.get("table_type") == "ibis":
            raise DeprecationWarning(
                "Calling `.table(..., table_type='ibis') is deprecated. Instead, call"
                " `.table('foo').to_ibis()` to create a `dlt.Relation` and then retrieve the"
                " Ibis Table."
            )

        if load_ids:
            return dlt.Relation(dataset=self, table_name=table_name).from_loads(load_ids)
        else:
            return dlt.Relation(dataset=self, table_name=table_name)

    def loads_table(self) -> dlt.Relation:
        """Get `_dlt_loads` table from the dataset."""
        return dlt.Relation(dataset=self, table_name=self.schema.loads_table_name)

    def load_ids(self) -> list[str]:
        """Retrieved the list of load ids for this dataset.

        This queries the `_dlt_loads` table on the destination and filters
        the `schema_name` columns based on the current dataset.
        """
        return _get_load_ids(self)

    def latest_load_id(self) -> Optional[str]:
        """Retrieved the latest load id for this dataset.

        This is the max value of the `load_id` column in the `_dlt_loads` table
        on the destination when filtering the `schema_name` columns based on the current dataset.
        """
        return _get_latest_load_id(self)

    def row_counts(
        self,
        *,
        data_tables: bool = True,
        dlt_tables: bool = False,
        table_names: Optional[list[str]] = None,
        load_id: Optional[str] = None,
    ) -> dlt.Relation:
        """Create a `dlt.Relation` with the query to get the row counts of all tables in the dataset.

        Args:
            data_tables (bool): Whether to include data tables. Defaults to True.
            dlt_tables (bool): Whether to include dlt tables. Defaults to False.
            table_names (Optional[list[str]]): The names of the tables to include. Defaults to None. Will override data_tables and dlt_tables if set
            load_id (Optional[str]): If set, only count rows associated with a given load id. Will exclude tables that do not have a load id.
        Returns:
            dlt.Relation: Relation for the query that computes the requested row count.
        """

        selected_tables = table_names or []
        if not selected_tables:
            if data_tables:
                selected_tables += self.schema.data_table_names(seen_data_only=True)
            if dlt_tables:
                selected_tables += self.schema.dlt_table_names()

        # filter tables so only ones with dlt_load_id column are included
        dlt_load_id_col = None
        if load_id:
            dlt_load_id_col = self.schema.naming.normalize_identifier(C_DLT_LOAD_ID)
            selected_tables = [
                table
                for table in selected_tables
                if dlt_load_id_col in self.schema.tables[table]["columns"].keys()
            ]

        union_all_expr = None

        for table_name in selected_tables:
            counts_expr = build_row_counts_expr(
                table_name=table_name,
                dlt_load_id_col=dlt_load_id_col,
                load_id=load_id,
            )
            if union_all_expr is None:
                union_all_expr = counts_expr
            else:
                union_all_expr = union_all_expr.union(counts_expr, distinct=False)

        return self.query(query=union_all_expr)

    def __getitem__(self, table_name: str) -> dlt.Relation:
        """Get a `dlt.Relation` for a table via dictionary notation.

        This proxies `Dataset.table()`.
        """
        if table_name not in self.tables:
            raise KeyError(
                f"Table `{table_name}` not found on dataset. Available tables: `{self.tables}`"
            )
        try:
            return self.table(table_name)
        # TODO: expect TableNotFound in the future
        except ValueError as exc:
            raise KeyError(table_name, str(exc))

    # TODO this creates all sorts of bugs in dynamic scenarios where `getattr()` is used
    # Table names that collide with `Dataset` methods will shadow attributes.
    # You get much weaker IDE autocompletion than `.__getitem__()` or `.table()`
    # Potential fix:
    # super().__getattr__(name); if None: or try/except: return self.table()
    def __getattr__(self, name: str) -> Any:
        """Get a `dlt.Relation` for a table via dictionary notation.

        This proxies `Dataset.table()`.
        """
        try:
            return self.table(name)
        # TODO: expect TableNotFound in the future
        except ValueError as exc:
            raise AttributeError(name, str(exc))

    def __enter__(self) -> Self:
        """Context manager to keep the connection to the destination open between queries"""
        assert (
            not self._opened_sql_client
        ), "context manager can't be used when sql client is initialized"
        # return sql_client wrapped so it will not call close on __exit__ as dataset is managing connections
        # use internal class

        # create a new sql client
        self._opened_sql_client = get_dataset_sql_client(self)

        NoCloseClient = type(
            "NoCloseClient",
            (self._opened_sql_client.__class__,),
            {
                "__exit__": (
                    lambda self, exc_type, exc_val, exc_tb: None
                )  # No-operation: do not close the connection
            },
        )

        self._opened_sql_client.__class__ = NoCloseClient

        self._opened_sql_client.open_connection()
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        """Context manager to keep the connection to the destination open between queries"""
        self._opened_sql_client.close_connection()
        self._opened_sql_client = None

    def __repr__(self) -> str:
        # schema may not be set
        kwargs = {
            "dataset_name": self.dataset_name,
            "destination": repr(self._destination),
            "schema": repr(self.schema) if self.schema else None,
        }
        return simple_repr("dlt.dataset", **without_none(kwargs))

    def __str__(self) -> str:
        # obtain schema. this eventually loads it from destination (if only name was provided)
        _client = self.destination_client
        # str(config) provides displayable physical location
        destination_info = f"{self._destination.destination_name}[{str(_client.config)}]"
        # new schema is created if dataset is not found or schema.name is not materialized yet
        if self.schema.is_new:
            schema_info = "\nDataset is not available at the destination.\n"
        else:
            schema_info = f"with tables in dlt schema `{self.schema.name}`:\n"

        msg = f"Dataset `{self.dataset_name}` at `{destination_info}` {schema_info}"
        if self.schema:
            tables = [name for name in self.schema.data_table_names()]
            if tables:
                msg += ", ".join(tables)
            else:
                msg += "No data tables found in schema."
        return msg


def dataset(
    destination: TDestinationReferenceArg,
    dataset_name: str,
    schema: Union[Schema, str, None] = None,
) -> Dataset:
    return Dataset(destination, dataset_name, schema)


def get_dataset_destination_client(dataset: dlt.Dataset) -> JobClientBase:
    return get_destination_clients(
        schema=dataset.schema,
        destination=dataset._destination,
        destination_dataset_name=dataset.dataset_name,
    )[0]


def get_dataset_sql_client(dataset: dlt.Dataset) -> SqlClientBase[Any]:
    client = get_dataset_destination_client(dataset)
    if isinstance(client, WithSqlClient):
        return client.sql_client
    else:
        raise SqlClientNotAvailable("dataset", dataset.dataset_name, client.config.destination_type)


def is_same_physical_destination(dataset1: dlt.Dataset, dataset2: dlt.Dataset) -> bool:
    """Check if both datasets are at the same physical destination.

    This is done by comparing the fingerprint of both destination configs. There
    are potential false positive if two different config give access to the same destination.
    """
    return str(dataset1.destination_client.config) == str(dataset2.destination_client.config)


def _get_dataset_schema_from_destination_using_schema_name(
    dataset: dlt.Dataset, schema_name: str
) -> Optional[dlt.Schema]:
    schema = None
    with get_destination_clients(
        schema=dlt.Schema(schema_name),
        destination=dataset._destination,
        destination_dataset_name=dataset.dataset_name,
    )[0] as client:
        if isinstance(client, WithStateSync):
            stored_schema = client.get_stored_schema(schema_name)
            if stored_schema:
                schema = dlt.Schema.from_stored_schema(json.loads(stored_schema.schema))
            else:
                schema = dlt.Schema(schema_name)

    return schema


def _get_dataset_schema_from_dataset_dlt_tables(
    dataset: dlt.Dataset,
) -> Optional[dlt.Schema]:
    """Resolves schema from destination. First tries pipeline state to get default_schema_name,
    then falls back to most recently stored schema."""
    with get_destination_clients(
        schema=dlt.Schema(dataset.dataset_name),
        destination=dataset._destination,
        destination_dataset_name=dataset.dataset_name,
    )[0] as client:
        if isinstance(client, WithStateSync):
            # try to resolve via pipeline state for correct default schema
            pipeline_name = dataset._pipeline_name
            if pipeline_name:
                try:
                    stored_state = client.get_stored_state(pipeline_name)
                except DestinationUndefinedEntity:
                    # state tables may not exist if pipeline never ran
                    stored_state = None
                if stored_state:
                    state = decompress_state(stored_state.state)
                    schema_name = state.get("default_schema_name")
                    if schema_name:
                        stored_schema = client.get_stored_schema(schema_name)
                        if stored_schema:
                            return dlt.Schema.from_stored_schema(json.loads(stored_schema.schema))
            # fall back to most recently stored schema
            stored_schema = client.get_stored_schema()
            if stored_schema:
                return dlt.Schema.from_stored_schema(json.loads(stored_schema.schema))
    return None


def _get_load_ids(dataset: dlt.Dataset) -> list[str]:
    """Get a list of load ids associated with the dataset."""
    loads_table = dataset.loads_table()
    query = (
        loads_table
        # need to filter out rare case where a dataset includes multiple `dlt.Schema`
        # this mechanism is currently used by data quality metrics and checks
        .where(dataset.schema.naming.normalize_identifier("schema_name"), "eq", dataset.schema.name)
        .select(dataset.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID))
        .order_by(dataset.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID), "asc")
    )
    load_ids: list[str] = [load_id[0] for load_id in query.fetchall()]
    return load_ids


def _get_latest_load_id(dataset: dlt.Dataset) -> Optional[str]:
    """Get the latest load id associated with the dataset."""
    loads_table = dataset.loads_table()
    query = (
        loads_table
        # need to filter out rare case where a dataset includes multiple `dlt.Schema`
        # this mechanism is currently used by data quality metrics and checks
        .where(dataset.schema.naming.normalize_identifier("schema_name"), "eq", dataset.schema.name)
        .select(dataset.schema.naming.normalize_identifier(C_DLT_LOADS_TABLE_LOAD_ID))
        .max()
    )
    load_id = query.fetchone()
    return load_id[0] if load_id else None  # type: ignore[no-any-return]



import contextlib
import functools
import threading
import time
from typing import Any, Callable, Dict, Optional

from dlt.common.configuration import resolve_configuration
from dlt.common.configuration.container import Container
from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext
from dlt.common.runtime.anon_tracker import track
from dlt.common.runtime.telemetry import _TELEMETRY_STARTED, start_telemetry
from dlt.common.typing import TFun
from dlt.common.utils import digest128

# serialize context reloads for concurrent MCP tool calls
_context_reload_lock = threading.Lock()


def _get_agent_info() -> Optional[Dict[str, str]]:
    """Extract connecting client metadata from the MCP session.

    Reads `clientInfo` sent by the MCP client during the `initialize`
    handshake.  The session id is hashed so the raw value is never stored.

    Returns:
        A dict with `name`, `version`, and `session_hash` when a
        session is active, or `None` outside of an MCP request.
    """
    try:
        from fastmcp.server.context import _current_context

        ctx = _current_context.get(None)
        if ctx is None:
            return None
        session = ctx.session
        client_params = session.client_params
        if client_params is None:
            return None
        info: Dict[str, str] = {
            "name": client_params.clientInfo.name,
            "version": client_params.clientInfo.version,
        }
        session_id = ctx.session_id
        if session_id:
            info["session_hash"] = digest128(session_id)
        return info
    except Exception:
        return None


def with_mcp_tool_telemetry(**extra_props: Any) -> Callable[[TFun], TFun]:
    """Telemetry decorator for MCP tool functions.

    Emits an anonymous `mcp_tool_{name}` telemetry event after each tool
    invocation.  Before executing the tool the decorator fully reloads the
    runtime context from disk (workspace marker, profile pin, secrets /
    config files) so that every MCP call sees the latest on-disk state.
    Reload and execution are serialized under a module-level lock.

    `agent_info` is `None` when the session metadata is not available
    (e.g. during tests that call tool functions directly).

    Args:
        **extra_props: literal key-value pairs added to the telemetry event
            properties.

    Returns:
        A decorator that wraps the tool function with telemetry tracking and
        context reload.
    """

    def decorator(fn: TFun) -> TFun:
        event_name = "tool_" + fn.__name__

        @functools.wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            props: Dict[str, Any] = {"mcp_primitive": "tool"}
            props.update(extra_props)
            props["agent_info"] = _get_agent_info()
            start_ts = time.time()

            def _track(success: bool) -> None:
                with contextlib.suppress(Exception):
                    props["elapsed"] = time.time() - start_ts
                    props["success"] = success
                    if not _TELEMETRY_STARTED:
                        c = resolve_configuration(RuntimeConfiguration())
                        start_telemetry(c)
                    track("mcp", event_name, props)

            with _context_reload_lock:
                ctx_plug = Container()[PluggableRunContext]
                ctx_plug.reload(ctx_plug.context.run_dir)

            try:
                rv = fn(*args, **kwargs)
                _track(True)
                return rv
            except Exception:
                _track(False)
                raise

        return wrapper  # type: ignore[return-value]

    return decorator



import base64
import hashlib
import binascii
from copy import copy
from typing import List, Tuple, Mapping

from dlt.common.json import json
from dlt.common.typing import DictStrAny
from dlt.common.utils import compressed_b64decode, compressed_b64encode
from dlt.common.typing import TypedDict


class TVersionedState(TypedDict, total=False):
    _state_version: int
    _version_hash: str
    _state_engine_version: int


def generate_state_version_hash(state: TVersionedState, exclude_attrs: List[str] = None) -> str:
    # generates hash out of stored schema content, excluding hash itself, version and local state
    state_copy = copy(state)
    exclude_attrs = exclude_attrs or []
    exclude_attrs.extend(["_state_version", "_state_engine_version", "_version_hash"])
    for attr in exclude_attrs:
        state_copy.pop(attr, None)  # type: ignore
    content = json.typed_dumpb(state_copy, sort_keys=True)
    h = hashlib.sha3_256(content)
    return base64.b64encode(h.digest()).decode("ascii")


def bump_state_version_if_modified(
    state: TVersionedState, exclude_attrs: List[str] = None
) -> Tuple[int, str, str]:
    """Bumps the `state` version and version hash if content modified, returns (new version, new hash, old hash) tuple"""
    hash_ = generate_state_version_hash(state, exclude_attrs)
    previous_hash = state.get("_version_hash")
    if not previous_hash:
        # if hash was not set, set it without bumping the version, that's the initial state
        pass
    elif hash_ != previous_hash:
        state["_state_version"] += 1

    state["_version_hash"] = hash_
    return state["_state_version"], hash_, previous_hash


def default_versioned_state() -> TVersionedState:
    return {"_state_version": 0, "_state_engine_version": 1}


def json_encode_state(state: TVersionedState) -> str:
    return json.typed_dumps(state)


def json_decode_state(state_str: str) -> DictStrAny:
    return json.typed_loads(state_str)  # type: ignore[no-any-return]


def compress_state(state: TVersionedState) -> str:
    return compressed_b64encode(json.typed_dumpb(state))


def decompress_state(state_str: str) -> DictStrAny:
    try:
        state_bytes = compressed_b64decode(state_str)
    except binascii.Error:
        return json.typed_loads(state_str)  # type: ignore[no-any-return]
    else:
        return json.typed_loadb(state_bytes)  # type: ignore[no-any-return]



from typing import Dict, List, Set

from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.schema import Schema
from dlt.common.schema.typing import (
    TPartialTableSchema,
    TTableSchema,
    TSchemaUpdate,
    TColumnSchemaBase,
)
from dlt.common.schema.utils import (
    ensure_compatible_tables,
    find_incomplete_columns,
    get_first_column_name_with_prop,
    has_table_seen_data,
    is_complete_column,
    is_nested_table,
)
from dlt.common.schema.exceptions import UnboundColumnException, UnboundColumnWithoutTypeException
from dlt.common import logger


def validate_and_update_schema(schema: Schema, schema_updates: List[TSchemaUpdate]) -> None:
    """Updates `schema` tables with partial tables in `schema_updates`"""
    for schema_update in schema_updates:
        for table_name, table_updates in schema_update.items():
            logger.info(f"Updating schema for table {table_name} with {len(table_updates)} deltas")
            for partial_table in table_updates:
                # ensure updates will pass
                if existing_table := schema.tables.get(partial_table["name"]):
                    ensure_compatible_tables(schema.name, existing_table, partial_table)

            for partial_table in table_updates:
                verify_partial_table(schema, partial_table)

            for partial_table in table_updates:
                # merge columns where we expect identifiers to be normalized
                schema.update_table(partial_table, normalize_identifiers=False)


def verify_partial_table(schema: Schema, partial_table: TPartialTableSchema) -> None:
    """Verifies partial tables before they are merged with existing tables in schema

    1. warns if root_key is added to existing table that has seen data (this will most likely fail)
    """
    table_name = partial_table["name"]  # names are already normalized
    if existing_table := schema.tables.get(table_name):
        if has_table_seen_data(existing_table):
            # we are adding to a table materialized in the destination
            root_key_col = get_first_column_name_with_prop(partial_table, "root_key")
            if root_key_col:
                logger.warning(
                    f"You are adding a root_key in column {root_key_col} to an already existing"
                    f" table {table_name}. "
                    "This will most likely fail due to NOT NULL constraint. Typically this happens"
                    " when you switch "
                    "write disposition from append/replace to merge - in that case dlt will create"
                    " required linking "
                    "information automatically. You can fix your table, drop it or disable root key"
                    " propagation. "
                    "See details here:"
                    " https://dlthub.com/docs/general-usage/merge-loading#switch-from-appendreplace-to-merge"
                )


def reconcile_null_only_columns(
    schema: Schema, null_only_columns: Dict[str, Set[str]]
) -> Dict[str, List[TColumnSchemaBase]]:
    """Reconcile null-only column names against the schema. Returns a mapping of table name
    to column bases for columns that still have no type after normalization.
    """
    result: Dict[str, List[TColumnSchemaBase]] = {}
    for table_name, null_cols in null_only_columns.items():
        table = schema.tables.get(table_name)
        if not table:
            continue
        still_null: List[TColumnSchemaBase] = []
        for col_name in sorted(null_cols):
            col = table["columns"].get(col_name)
            if col is None or not is_complete_column(col):
                still_null.append(TColumnSchemaBase(name=col_name))
        if still_null:
            result[table_name] = still_null
    return result


def verify_normalized_table(
    schema: Schema,
    table: TTableSchema,
    capabilities: DestinationCapabilitiesContext,
    null_only_columns: List[TColumnSchemaBase] = None,
) -> None:
    """Verify `table` schema is valid for next stage after normalization. Only tables that have
    seen data are verified. Verification happens before seen-data flag is set so new tables
    can be detected.

    1. Log warning if any incomplete nullable columns are in any data tables
    2. Log warning for null-only columns that never received data
    3. Raise `UnboundColumnException` on incomplete non-nullable columns (e.g. missing merge/primary key)
    4. Log warning if table format is not supported by destination capabilities
    """
    null_only_names = {c["name"] for c in (null_only_columns or [])}
    incomplete_nullable_user: List[TColumnSchemaBase] = []

    for column, nullable in find_incomplete_columns(table):
        if nullable:
            if column["name"] not in null_only_names:
                incomplete_nullable_user.append(column)
        else:
            raise UnboundColumnException(schema.name, table["name"], [column])

    if null_only_columns:
        logger.warning(
            str(UnboundColumnWithoutTypeException(schema.name, table["name"], null_only_columns))
        )

    if incomplete_nullable_user:
        logger.warning(
            str(UnboundColumnException(schema.name, table["name"], incomplete_nullable_user))
        )

    # TODO: 3. raise if we detect name conflict for SCD2 columns
    # until we track data per column we won't be able to implement this
    # if resolve_merge_strategy(schema.tables, table, capabilities) == "scd2":
    #     for validity_column_name in get_validity_column_names(table):
    #         if validity_column_name in item.keys():
    #             raise ColumnNameConflictException(
    #                 schema_name,
    #                 "Found column in data item with same name as validity column"
    #                 f' "{validity_column_name}".',
    #             )

    supported_table_formats = capabilities.supported_table_formats or []
    if "table_format" in table and table["table_format"] not in supported_table_formats:
        logger.warning(
            "Destination does not support the configured `table_format` value "
            f"`{table['table_format']}` for table `{table['name']}`. "
            "The setting will probably be ignored."
        )

    parent_key = get_first_column_name_with_prop(table, "parent_key")
    if parent_key and not is_nested_table(table):
        logger.warning(
            f"Table {table['name']} has parent_key on column {parent_key} but no corresponding"
            " `parent` table hint to refer to parent table.Such table is not considered a nested"
            " table and relational normalizer will not generate linking data. The most probable"
            " cause is manual modification of the dlt schema for the table. The most probable"
            f" outcome will be NULL violation during the load process on {parent_key}."
        )



from typing import Dict, Any, Literal, Set

from dlt.common.exceptions import ValueErrorWithKnownValues
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.typing import TColumnNames, get_args
from dlt.extract import DltResource, resource as make_resource
from dlt.destinations.utils import get_resource_for_adapter

TTokenizationTMethod = Literal["word", "lowercase", "whitespace", "field"]
TOKENIZATION_METHODS: Set[TTokenizationTMethod] = set(get_args(TTokenizationTMethod))
TTokenizationSetting = Dict[str, TTokenizationTMethod]
"""Maps column names to tokenization types supported by Weaviate"""

VECTORIZE_HINT = "x-weaviate-vectorize"
TOKENIZATION_HINT = "x-weaviate-tokenization"


def weaviate_adapter(
    data: Any,
    vectorize: TColumnNames = None,
    tokenization: TTokenizationSetting = None,
) -> DltResource:
    """Prepares data for the Weaviate destination by specifying which columns
    should be vectorized and which tokenization method to use.

    Vectorization is done by Weaviate's vectorizer modules. The vectorizer module
    can be configured in dlt configuration file under
    `[destination.weaviate.vectorizer]` and `[destination.weaviate.module_config]`.
    The default vectorizer module is `text2vec-openai`. See also:
    https://weaviate.io/developers/weaviate/modules/retriever-vectorizer-modules

    Args:
        data (Any): The data to be transformed. It can be raw data or an instance
            of DltResource. If raw data, the function wraps it into a DltResource
            object.
        vectorize (TColumnNames, optional): Specifies columns that should be
            vectorized. Can be a single column name as a string or a list of
            column names.
        tokenization (TTokenizationSetting, optional): A dictionary mapping column
            names to tokenization methods supported by Weaviate. The tokenization
            methods are one of the values in `TOKENIZATION_METHODS`:
            - 'word',
            - 'lowercase',
            - 'whitespace',
            - 'field'.

    Returns:
        DltResource: A resource with applied Weaviate-specific hints.

    Raises:
        ValueError: If input for `vectorize` or `tokenization` is invalid
            or neither is specified.

    Examples:
        >>> data = [{"name": "Alice", "description": "Software developer"}]
        >>> weaviate_adapter(data, vectorize="description", tokenization={"description": "word"})
        [DltResource with hints applied]
    """
    resource = get_resource_for_adapter(data)

    column_hints: TTableSchemaColumns = {}
    if vectorize:
        if isinstance(vectorize, str):
            vectorize = [vectorize]
        if not isinstance(vectorize, list):
            raise ValueError(
                "`vectorize` must be a list of column names or a single column name as a string"
            )
        # create weaviate-specific vectorize hints
        for column_name in vectorize:
            column_hints[column_name] = {
                "name": column_name,
                VECTORIZE_HINT: True,  # type: ignore
            }

    if tokenization:
        for column_name, method in tokenization.items():
            if method not in TOKENIZATION_METHODS:
                raise ValueErrorWithKnownValues("method", method, TOKENIZATION_METHODS)

            if column_name in column_hints:
                column_hints[column_name][TOKENIZATION_HINT] = method  # type: ignore
            else:
                column_hints[column_name] = {
                    "name": column_name,
                    TOKENIZATION_HINT: method,  # type: ignore
                }

    # this makes sure that {} as column_hints never gets into apply_hints (that would reset existing columns)
    if not column_hints:
        raise ValueError("Either 'vectorize' or 'tokenization' must be specified.")
    else:
        resource.apply_hints(columns=column_hints)

    return resource



import posixpath
from typing import Any, Dict, List, Literal, Optional, Tuple

from dlt.common.json import json
from dlt.common.configuration.specs import GcpCredentials, GcpOAuthCredentials
from dlt.common.exceptions import MissingDependencyException

from fsspec.spec import AbstractFileSystem, AbstractBufferedFile

try:
    from googleapiclient.discovery import build
    from googleapiclient.errors import HttpError
except ModuleNotFoundError:
    raise MissingDependencyException("GoogleDriveFileSystem", ["google-api-python-client"])

try:
    from google.auth.credentials import AnonymousCredentials
except ModuleNotFoundError:
    raise MissingDependencyException(
        "GoogleDriveFileSystem", ["google-auth-httplib2 google-auth-oauthlib"]
    )

SCOPES = {
    "full_control": "https://www.googleapis.com/auth/drive",
    "read_only": "https://www.googleapis.com/auth/drive.readonly",
}


DIR_MIME_TYPE = "application/vnd.google-apps.folder"
FILE_INFO_FIELDS = ",".join(
    [
        "name",
        "id",
        "size",
        "description",
        "mimeType",
        "version",
        "createdTime",
        "modifiedTime",
        "capabilities",
    ]
)


DEFAULT_BLOCK_SIZE = 5 * 2**20
# maps (parent id, child_name) into file id
FILE_ID_CACHE: Dict[Tuple[str, str], str] = {}


class GoogleDriveFileSystem(AbstractFileSystem):
    protocol = "gdrive"
    root_marker = ""

    def __init__(
        self,
        credentials: GcpCredentials = None,
        trash_delete: bool = True,
        access: Optional[Literal["full_control", "read_only"]] = "full_control",
        spaces: Optional[Literal["drive", "appDataFolder", "photos"]] = "drive",
        **kwargs: Any,
    ):
        """Google Drive as a file-system.

        The gdrive url has following format: gdrive://<root_file_id>/<file_path>
        Where <root_file_id> is a file id of the folder where the <file_path> is present.

        Google Drive provides consistency when file ids are used. Changes are reflected immediately.
        In case of listings (ls) the consistency is eventual. Changes are reflected with a delay.
        As ls is used to retrieve file id from file name, we will be unable to build consistent filesystem
        with google drive API. Use this with care.

        Based on original fsspec Google Drive implementation: https://github.com/fsspec/gdrivefs

        Args:
            credentials (GcpCredentials): Google Service credentials. If not provided, anonymous credentials
                are used
            trash_delete (bool): If True sends files to trash on rm. If False, deletes permanently.
                Note that permanent delete is not available for shared drives
            access (Optional[Literal["full_control", "read_only"]]):
                One of "full_control", "read_only".
            spaces (Optional[Literal["drive", "appDataFolder", "photos"]]):
                Category of files to search, can be 'drive', 'appDataFolder' and 'photos'.
                Of these, only the first is general.
            **kwargs:
                Passed to the parent.
        """
        super().__init__(**kwargs)
        self.trash_delete = trash_delete
        self.access = access
        self.scopes = [SCOPES[access]]
        self.credentials = credentials
        self.spaces = spaces
        self.connect()

    def connect(self) -> None:
        """Connect to Google Drive."""

        if self.credentials:
            if isinstance(self.credentials, GcpOAuthCredentials):
                self.credentials.auth(self.scopes)
            cred = self.credentials.to_native_credentials()
        else:
            cred = AnonymousCredentials()

        srv = build("drive", "v3", credentials=cred)
        self.service = srv.files()

    def mkdir(self, path: str, create_parents: Optional[bool] = True) -> None:
        """Create a directory.

        Args:
            path (str): The directory to create.
            create_parents (Optional[bool]):
                Whether to create parent directories if they don't exist.
                Defaults to True.
        """
        # if there are more than two components, create parents (first component is root id, second - path being created)
        if create_parents and len(path.split("/")) > 2:
            self.makedirs(self._parent(path), exist_ok=True)
        parent_id = self.path_to_file_id(self._parent(path))
        name = path.rstrip("/").rsplit("/", 1)[-1]
        meta = {
            "name": name,
            "mimeType": DIR_MIME_TYPE,
            "parents": [parent_id],
        }
        file = self.service.create(body=meta, supportsAllDrives=True).execute()
        # cache the new dir
        FILE_ID_CACHE[(parent_id, name)] = file["id"]
        try:
            self.invalidate_cache(parent_id)
        except Exception:
            # allow for parent to not exist as valid path
            pass

    def makedirs(self, path: str, exist_ok: Optional[bool] = True) -> None:
        """Create a directory and all its parent components.

        Args:
            path (str): The directory to create.
            exist_ok (Optional[bool]): Whether to raise an error if the directory already exists.
                Defaults to True.
        """
        if self.isdir(path):
            if exist_ok:
                return
            else:
                raise FileExistsError(path)
        # if there are more than two components, create parents (first component is root id, second - path being created)
        if len(path.split("/")) > 2:
            self.makedirs(self._parent(path), exist_ok=True)
        self.mkdir(path, create_parents=False)

    def _is_path_root_id(self, path: str) -> bool:
        """Checks if path is root id"""
        return len(path.split("/")) == 1

    def _delete(self, file_id: str) -> None:
        """Delete a file.

        Args:
            file_id (str): The ID of the file to delete.
        """
        self.service.delete(fileId=file_id, supportsAllDrives=True).execute()

    def _trash(self, file_id: str) -> None:
        """Sends file to trash.

        Args:
            file_id (str): The ID of the file to trash.
        """
        file_metadata = {"trashed": True}
        self.service.update(fileId=file_id, supportsAllDrives=True, body=file_metadata).execute()

    def rm(
        self, path: str, recursive: Optional[bool] = True, maxdepth: Optional[int] = None
    ) -> None:
        """Remove files or directories.

        Args:
            path (str): The file or directory to remove.
            recursive (Optional[bool]): Whether to remove directories recursively.
                Defaults to True.
            maxdepth (Optional[int]): The maximum depth to remove directories.
        """
        if recursive is False and self.isdir(path) and self.ls(path):
            raise ValueError("Attempt to delete non-empty folder")
        file_id = self.path_to_file_id(path)
        if self.trash_delete:
            self._trash(file_id)
        else:
            self._delete(file_id)
        self.invalidate_cache(file_id)
        try:
            if not self._is_path_root_id(path):
                parent_id = self.path_to_file_id(self._parent(path))
                # drop file id if exists
                FILE_ID_CACHE.pop((parent_id, path.split("/")[-1]), None)
                self.invalidate_cache(parent_id)
        except Exception:
            # allow for parent to not exist as valid path
            pass

    def invalidate_cache(self, file_id: str) -> None:
        # remove listing from cache by file id
        self.dircache.pop(file_id, None)

    def rmdir(self, path: str) -> None:
        """Remove a directory.

        Args:
            path (str): The directory to remove.
        """
        if not self.isdir(path):
            raise ValueError("Path is not a directory")
        self.rm(path, recursive=False)

    def info(self, path: str, **kwargs: Any) -> Dict[str, Any]:
        file_id = self.path_to_file_id(path)
        return self._info_by_id(file_id, self._parent(path))

    def _info_by_id(self, file_id: str, path_prefix: Optional[str] = None) -> Dict[str, Any]:
        response = self.service.get(
            fileId=file_id,
            fields=FILE_INFO_FIELDS,
            supportsAllDrives=True,
        ).execute()
        return self._file_info_from_response(response, path_prefix)

    def export(self, path: str, mime_type: str) -> Any:
        """Convert a Google-native file to other format and download

        mime_type is something like "text/plain"
        """
        file_id = self.path_to_file_id(path)
        return self.service.export(
            fileId=file_id, mimeType=mime_type, supportsAllDrives=True
        ).execute()

    def ls(self, path: str, detail: Optional[bool] = False, refresh: Optional[bool] = False) -> Any:
        """List files in a directory.

        Args:
            path (str): The directory to list.
            detail (Optional[bool]): Whether to return detailed file information.
                Defaults to False.

        Returns:
            Any: Files in the directory data.
        """
        file_id = self.path_to_file_id(path)
        files = self._list_directory_by_id(file_id, dir_path=path)

        if detail:
            return files
        else:
            return [f["name"] for f in files]

    def _list_directory_by_id(
        self, dir_file_id: str, dir_path: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """List files in a directory by ID.

        Args:
            dir_file_id (str): The ID of the directory to list.
            path_prefix (Optional[str]): The path prefix to use.

        Returns:
            List[Dict[str, Any]]: A list of files in the directory.
        """
        # use file_id for caching
        files = self.dircache.get(dir_file_id)
        if files is not None:
            return files  # type: ignore

        all_files = []
        page_token = None
        all_fields = "nextPageToken, files(%s)" % FILE_INFO_FIELDS
        query = f"'{dir_file_id}' in parents "

        while True:
            response = self.service.list(
                q=query,
                spaces=self.spaces,
                fields=all_fields,
                pageToken=page_token,
                includeItemsFromAllDrives=True,
                supportsAllDrives=True,
                corpora="allDrives",
            ).execute()

            for f in response.get("files", []):
                all_files.append(self._file_info_from_response(f, dir_path))

            page_token = response.get("nextPageToken", None)
            if page_token is None:
                break

        self.dircache[dir_file_id] = all_files
        return all_files

    def path_to_file_id(
        self, path: str, parent_id: Optional[str] = None, parent_path: str = ""
    ) -> str:
        """Get the file ID from a path.

        Args:
            path (str): The path to get the file ID from.
            parent_id (Optional[str]): The parent directory id to search.
            parent_path (Optional[str]): Path corresponding to parent id

        Returns:
            str: The file ID.
        """
        items = path.strip("/").split("/", 3)
        if not parent_id:
            # must have root id
            if len(items) == 0:
                raise ValueError(
                    "Google drive path must start with folder root id ie."
                    " gdrive://15eC3e5MNew2XAIefWNlG8VlEa0ISnnaG/..."
                )
            if len(items) == 1:
                return items[0]  #  only root id present
            parent_id, file_name = items[:2]
            descendants = items[2:]
            parent_path = parent_id
        else:
            if len(items) == 0 or len(path) == 0:
                return parent_id
            file_name = items[0]
            descendants = items[1:]

        # use cached file ids
        top_file_id = FILE_ID_CACHE.get((parent_id, file_name))
        if not top_file_id:
            top_file_id = self._find_file_id_in_dir(file_name, parent_id, parent_path)
            FILE_ID_CACHE[(parent_id, file_name)] = top_file_id
        if not descendants:
            return top_file_id
        else:
            sub_path = posixpath.join(*descendants)
            return self.path_to_file_id(
                sub_path, parent_id=top_file_id, parent_path=posixpath.join(parent_path, file_name)
            )

    def _find_file_id_in_dir(self, file_name: str, dir_file_id: str, dir_path: str) -> Any:
        """Get the file ID of a file with a given name in a directory.

        Args:
            child_name (str): The name of the child to get the file ID of.
            dir_file_id (str): The ID of the directory to search.
            dir_path (str): Path corresponding to dir_file_id

        Returns:
            str: The file ID of the file_name.
        """
        all_children = self._list_directory_by_id(dir_file_id, dir_path=dir_path)
        possible_children = []
        for child in all_children:
            if child["name"].strip("/").split("/")[-1] == file_name:
                possible_children.append(child["id"])

        if len(possible_children) == 0:
            raise FileNotFoundError(f"Directory `{dir_file_id}` has no child named `{file_name}`")
        if len(possible_children) == 1:
            return possible_children[0]
        else:
            raise KeyError(
                f"Directory `{dir_file_id}` has more than one "
                f"child named `{file_name}`. Unable to resolve path to `file_id`."
            )

    def _open(self, path: str, mode: Optional[str] = "rb", **kwargs: Any) -> "GoogleDriveFile":
        """Open a file.

        Args:
            path (str): The file to open.
            mode (Optional[str]): The mode to open the file in.
                Defaults to "rb".
            **kwargs: Passed to the parent.

        Returns:
            GoogleDriveFile: The opened file.
        """
        return GoogleDriveFile(self, path, mode=mode, **kwargs)

    @staticmethod
    def _file_info_from_response(file: Dict[str, Any], path_prefix: str = None) -> Dict[str, Any]:
        """Create fsspec compatible file info"""
        ftype = "directory" if file.get("mimeType") == DIR_MIME_TYPE else "file"
        if path_prefix:
            name = posixpath.join(path_prefix, file["name"])
        else:
            name = file["name"]

        info = {"name": name, "size": int(file.get("size", 0)), "type": ftype}
        file.update(info)
        return file


class GoogleDriveFile(AbstractBufferedFile):
    def __init__(
        self,
        fs: GoogleDriveFileSystem,
        path: str,
        mode: Optional[str] = "rb",
        block_size: Optional[int] = DEFAULT_BLOCK_SIZE,
        autocommit: Optional[bool] = True,
        **kwargs: Any,
    ):
        """A Google Drive file.

        Args:
            fs (AbstractFileSystem): The file system to open the file from.
            path (str): The file to open.
            mode (Optional[str]): The mode to open the file in.
                Defaults to "rb".
            block_size (Optional[str]): The block size to use.
                Defaults to DEFAULT_BLOCK_SIZE.
            autocommit (Optional[bool]): Whether to automatically commit the file.
                Defaults to True.
            **kwargs: Passed to the parent.
        """
        self.root_file_id, _ = path.split("/", 1)
        super().__init__(fs, path, mode, block_size, autocommit=autocommit, **kwargs)
        self.fs = fs
        # try to get file_id
        try:
            self.file_id = fs.path_to_file_id(path)
        except FileNotFoundError:
            if "r" in mode:
                raise
            self.file_id = None
        self.parent_id: str = None
        self.location = None

    def _fetch_range(self, start: Optional[int] = None, end: Optional[int] = None) -> Any:
        """Read data from Google Drive.

        Args:
            start (Optional[int]): The start of the range to read.
            end (Optional[int]): The end of the range to read.

        Returns:
            Any: The data read from Google Drive.
        """

        if start is not None or end is not None:
            start = start or 0
            end = end or 0
            head = {"Range": "bytes=%i-%i" % (start, end - 1)}
        else:
            head = {}
        try:
            files_service = self.fs.service
            media_obj = files_service.get_media(fileId=self.file_id)
            media_obj.headers.update(head)
            data = media_obj.execute()
            return data
        except HttpError as e:
            # TODO : doc says server might send everything if range is outside
            if "not satisfiable" in str(e):
                return b""
            raise

    def _upload_chunk(self, final: Optional[bool] = False) -> bool:
        """Write one part of a multi-block file upload.

        Args:
            final (Optional[bool]): Whether to finalize the file.
                Defaults to False.

        Returns:
            bool: Whether the upload was successful.
        """
        self.buffer.seek(0)
        data = self.buffer.getvalue()
        head = {}
        length = len(data)
        if final and self.autocommit:
            if length:
                part = "%i-%i" % (self.offset, self.offset + length - 1)
                head["Content-Range"] = "bytes %s/%i" % (part, self.offset + length)
            else:
                # closing when buffer is empty
                head["Content-Range"] = "bytes */%i" % self.offset
                data = None
        else:
            head["Content-Range"] = "bytes %i-%i/*" % (self.offset, self.offset + length - 1)
        head.update({"Content-Type": "application/octet-stream", "Content-Length": str(length)})
        req = self.fs.service._http.request
        head, body = req(self.location, method="PUT", body=data, headers=head)
        status = int(head["status"])
        # TODO: raise an exception similar to google api wrapper
        assert status < 400, "Init upload failed"
        if status in [200, 201]:
            # server thinks we are finished, this should happen
            # only when closing
            file_meta = json.loads(body.decode())
            self.file_id = file_meta["id"]
            # file_parent_id = file_meta["parents"][0]
            file_name = file_meta["name"]
            FILE_ID_CACHE[(self.parent_id, file_name)] = self.file_id
            # TODO: invalidate listing cache
        elif "range" in head:
            assert status == 308
        else:
            raise IOError
        return True

    def commit(self) -> None:
        """If not auto-committing, finalize the file."""
        self.autocommit = True
        self._upload_chunk(final=True)

    def _initiate_upload(self) -> None:
        """Create a multi-upload."""
        parent_path = self.fs._parent(self.path)
        if parent_path == self.root_file_id:
            self.parent_id = self.root_file_id
        else:
            self.parent_id = self.fs.path_to_file_id(self.fs._parent(self.path))
        head = {"Content-Type": "application/json; charset=UTF-8"}
        query = "https://www.googleapis.com/upload/drive/v3/files"
        if self.file_id:
            query += "/" + self.file_id
            body = {}
        else:
            # TODO: infer mime type from extension
            body = {"name": self.path.rsplit("/", 1)[-1], "parents": [self.parent_id]}
        query += "?uploadType=resumable&supportsAllDrives=true"
        req = self.fs.service._http.request
        head, body = req(
            query,
            method="PATCH" if self.file_id else "POST",
            headers=head,
            body=json.dumps(body).encode("utf-8"),
        )
        assert int(head["status"]) < 400, "Init upload failed"
        self.location = head["location"]  # type: ignore

    def discard(self) -> None:
        """Cancel in-progress multi-upload."""
        if self.location is None:
            return
        # uid = re.findall("upload_id=([^&=?]+)", self.location)
        # # TODO: there's no gcfs property, I could never work
        # head, _ = self.gcsfs._call(
        #     "DELETE",
        #     "https://www.googleapis.com/upload/drive/v3/files",
        #     params={"uploadType": "resumable", "upload_id": uid},
        # )
        # assert int(head["status"]) < 400, "Cancel upload failed"



from typing import Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.typing import TSecretStrValue


@configspec
class DBTCloudConfiguration(BaseConfiguration):
    api_token: TSecretStrValue = ""

    account_id: Optional[str] = None
    job_id: Optional[str] = None
    project_id: Optional[str] = None
    environment_id: Optional[str] = None
    run_id: Optional[str] = None

    cause: str = "Triggered via API"
    git_sha: Optional[str] = None
    git_branch: Optional[str] = None
    schema_override: Optional[str] = None



from dataclasses import dataclass
from typing import Any, List, Dict, Union, Sequence, Optional, cast

from dlt.common.destination.typing import PreparedTableSchema
from dlt.destinations.utils import get_resource_for_adapter
from dlt.extract import DltResource

PARTITION_HINT = "x-iceberg-partition"
TABLE_PROPERTIES_HINT = "x-iceberg-table-properties"


@dataclass(frozen=True)
class PartitionSpec:
    source_column: str
    transform: str = "identity"
    param_value: Optional[int] = None
    partition_field: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        d: Dict[str, Any] = {
            "transform": self.transform,
            "source_column": self.source_column,
        }
        if self.partition_field:
            d["partition_field"] = self.partition_field
        if self.param_value is not None:
            d["param_value"] = self.param_value
        return d

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "PartitionSpec":
        return cls(
            source_column=d["source_column"],
            transform=d["transform"],
            param_value=d.get("param_value"),
            partition_field=d.get("partition_field"),
        )


class iceberg_partition:
    """Helper class with factory methods for creating partition specs."""

    @staticmethod
    def identity(column_name: str) -> PartitionSpec:
        """Create an identity partition on a column.

        Args:
            column_name: The name of the column to partition on

        Returns:
            A PartitionSpec for identity partitioning
        """
        return PartitionSpec(column_name, "identity")

    @staticmethod
    def year(column_name: str, partition_field_name: Optional[str] = None) -> PartitionSpec:
        """Create a year partition on a timestamp/date column.

        Args:
            column_name: The name of the column to partition on
            partition_field_name: Optional custom name for the partition field

        Returns:
            A PartitionSpec for year partitioning
        """
        return PartitionSpec(column_name, "year", partition_field=partition_field_name)

    @staticmethod
    def month(column_name: str, partition_field_name: Optional[str] = None) -> PartitionSpec:
        """Create a month partition on a timestamp/date column.

        Args:
            column_name: The name of the column to partition on
            partition_field_name: Optional custom name for the partition field

        Returns:
            A PartitionSpec for month partitioning
        """
        return PartitionSpec(column_name, "month", partition_field=partition_field_name)

    @staticmethod
    def day(column_name: str, partition_field_name: Optional[str] = None) -> PartitionSpec:
        """Create a day partition on a timestamp/date column.

        Args:
            column_name: The name of the column to partition on
            partition_field_name: Optional custom name for the partition field

        Returns:
            A PartitionSpec for day partitioning
        """
        return PartitionSpec(column_name, "day", partition_field=partition_field_name)

    @staticmethod
    def hour(column_name: str, partition_field_name: Optional[str] = None) -> PartitionSpec:
        """Create an hour partition on a timestamp column.

        Args:
            column_name: The name of the column to partition on
            partition_field_name: Optional custom name for the partition field

        Returns:
            A PartitionSpec for hour partitioning
        """
        return PartitionSpec(column_name, "hour", partition_field=partition_field_name)

    @staticmethod
    def bucket(
        num_buckets: int, column_name: str, partition_field_name: Optional[str] = None
    ) -> PartitionSpec:
        """Create a bucket partition on a column.

        Args:
            num_buckets: The number of buckets to create
            column_name: The name of the column to partition on
            partition_field_name: Optional custom name for the partition field

        Returns:
            A PartitionSpec for bucket partitioning
        """
        return PartitionSpec(
            source_column=column_name,
            transform="bucket",
            param_value=num_buckets,
            partition_field=partition_field_name,
        )

    @staticmethod
    def truncate(
        width: int, column_name: str, partition_field_name: Optional[str] = None
    ) -> PartitionSpec:
        """Create a truncate partition on a string column.

        Args:
            width: The width to truncate to
            column_name: The name of the column to partition on
            partition_field_name: Optional custom name for the partition field

        Returns:
            A PartitionSpec for truncate partitioning
        """
        return PartitionSpec(
            source_column=column_name,
            transform="truncate",
            param_value=width,
            partition_field=partition_field_name,
        )


def iceberg_adapter(
    data: Any,
    partition: Union[str, PartitionSpec, Sequence[Union[str, PartitionSpec]]] = None,
    table_properties: Optional[Dict[str, str]] = None,
) -> DltResource:
    """Prepares data or a DltResource for loading into Apache Iceberg table.

    Takes raw data or an existing DltResource and configures it for Iceberg
    by defining partitioning strategies and/or table properties via the
    DltResource's hints.

    Args:
        data: The data to be transformed. This can be raw data (e.g., list of dicts)
            or an instance of `DltResource`. If raw data is provided, it will be
            encapsulated into a `DltResource` instance.
        partition: Defines how the Iceberg table should be partitioned.
            It accepts:
            - A single column name (string): Defaults to an identity transform.
            - A `PartitionSpec` object: Allows for detailed partition configuration,
              including transformation types (year, month, day, hour, bucket, truncate).
              Use the `iceberg_partition` helper class to create these specs.
            - A sequence of the above: To define multiple partition columns.
        table_properties: A dictionary of Iceberg table properties to set on the
            table at creation time. Keys and values must be strings. These correspond
            to Iceberg table properties such as ``write.format.default`` or
            ``write.target-file-size-bytes``. Properties are only applied when the
            table is first created.

    Returns:
        A `DltResource` instance configured with Iceberg-specific hints,
        ready for loading.

    Raises:
        ValueError: If neither `partition` nor `table_properties` is specified,
            or if an invalid partition transform is requested within a
            `PartitionSpec`, or if `table_properties` is not a dict with string
            keys and values.

    Examples:
        >>> data = [{"id": 1, "event_time": "2023-03-15T10:00:00Z", "category": "A"}]
        >>> resource = iceberg_adapter(
        ...     data,
        ...     partition=[
        ...         "category",  # Identity partition on category
        ...         iceberg_partition.year("event_time"),
        ...     ],
        ...     table_properties={"write.format.default": "parquet"},
        ... )
    """
    resource = get_resource_for_adapter(data)
    additional_table_hints: Dict[str, Any] = {}

    if partition:
        if isinstance(partition, (str, PartitionSpec)):
            partition = [partition]

        specs: List[PartitionSpec] = []
        for item in partition:
            if isinstance(item, PartitionSpec):
                specs.append(item)
            else:
                # item is the column name, use identity transform
                specs.append(iceberg_partition.identity(item))

        additional_table_hints[PARTITION_HINT] = [spec.to_dict() for spec in specs]

    if table_properties is not None:
        if not isinstance(table_properties, dict):
            raise ValueError("`table_properties` must be a dictionary.")
        for key, value in table_properties.items():
            if not isinstance(key, str):
                raise ValueError("Table property keys must be strings.")
            if not isinstance(value, str):
                raise ValueError(
                    f"Table property values must be strings. Got {type(value).__name__}"
                    f" for key '{key}'."
                )
        additional_table_hints[TABLE_PROPERTIES_HINT] = table_properties

    if additional_table_hints:
        resource.apply_hints(additional_table_hints=additional_table_hints)
    else:
        raise ValueError("At least one of `partition` or `table_properties` must be specified.")

    return resource


def parse_partition_hints(table_schema: PreparedTableSchema) -> List[PartitionSpec]:
    """Parse PARTITION_HINT from table schema into PartitionSpec list.

    Args:
        table_schema: dlt table schema containing partition hints

    Returns:
        List of PartitionSpec objects from hints, empty list if no hints found
    """
    partition_hints = cast(List[Dict[str, Any]], table_schema.get(PARTITION_HINT, []))
    specs = []
    for spec_data in partition_hints:
        spec = PartitionSpec.from_dict(spec_data)
        specs.append(spec)
    return specs


def create_identity_specs(column_names: List[str]) -> List[PartitionSpec]:
    """Create identity partition specs from column names.

    Args:
        column_names: List of column names to partition by identity

    Returns:
        List of PartitionSpec objects with identity transform
    """
    return [iceberg_partition.identity(column_name) for column_name in column_names]



from typing import Any, Dict, Tuple, Type, Union, TYPE_CHECKING, Optional

from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.data_writers.escape import escape_snowflake_identifier
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema.typing import TColumnSchema, TColumnType

from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.snowflake.configuration import (
    SnowflakeCredentials,
    SnowflakeClientConfiguration,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.snowflake.snowflake import SnowflakeClient


class SnowflakeTypeMapper(TypeMapperImpl):
    BIGINT_PRECISION = 19
    sct_to_unbound_dbt = {
        "json": "VARIANT",
        "text": "VARCHAR",
        "double": "FLOAT",
        "bool": "BOOLEAN",
        "date": "DATE",
        "timestamp": "TIMESTAMP_TZ",
        "bigint": f"NUMBER({BIGINT_PRECISION},0)",  # Snowflake has no integer types
        "binary": "BINARY",
        "time": "TIME",
        "decimal": "DECIMAL",
    }

    sct_to_dbt = {
        "text": "VARCHAR(%i)",
        "timestamp": "TIMESTAMP_TZ(%i)",
        "decimal": "NUMBER(%i,%i)",
        "time": "TIME(%i)",
        "wei": "NUMBER(%i,%i)",
    }

    dbt_to_sct = {
        "VARCHAR": "text",
        "FLOAT": "double",
        "BOOLEAN": "bool",
        "DATE": "date",
        "TIMESTAMP_TZ": "timestamp",
        "BINARY": "binary",
        "VARIANT": "json",
        "TIME": "time",
        "DECFLOAT": "decimal",
        "DECIMAL": "decimal",
    }

    def __init__(
        self,
        capabilities: DestinationCapabilitiesContext,
        use_decfloat: bool = False,
    ) -> None:
        super().__init__(capabilities)
        self.use_decfloat = use_decfloat

    def from_destination_type(
        self, db_type: str, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> TColumnType:
        if db_type == "NUMBER":
            if precision == self.BIGINT_PRECISION and scale == 0:
                return dict(data_type="bigint")
            elif (precision, scale) == self.capabilities.wei_precision:
                return dict(data_type="wei")
            return dict(data_type="decimal", precision=precision, scale=scale)
        if db_type == "TIMESTAMP_NTZ":
            return dict(data_type="timestamp", precision=precision, scale=scale, timezone=False)
        return super().from_destination_type(db_type, precision, scale)

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        timezone = column.get("timezone", True)
        precision = column.get("precision")

        if timezone and precision is None:
            # use lookup table for non-precision types
            return None

        timestamp = "TIMESTAMP_TZ" if timezone else "TIMESTAMP_NTZ"

        # append precision if specified and valid
        if precision is not None:
            if 0 <= precision <= self.capabilities.max_timestamp_precision:
                timestamp += f"({precision})"
            else:
                column_name = column["name"]
                table_name = table["name"]
                raise TerminalValueError(
                    f"Snowflake does not support `{precision=:}` for timestamp column"
                    f" `{column_name}` in table `{table_name}`"
                )

        return timestamp

    def decimal_precision(
        self, precision: Optional[int] = None, scale: Optional[int] = None
    ) -> Optional[Tuple[int, int]]:
        # when use_decfloat is enabled, unbound decimals map to DECFLOAT (no precision)
        if self.use_decfloat and precision is None and scale is None:
            return None
        return super().decimal_precision(precision, scale)

    def to_db_decimal_type(self, column: TColumnSchema) -> str:
        precision_tup = self.decimal_precision(column.get("precision"), column.get("scale"))
        if self.use_decfloat and precision_tup is None:
            return "DECFLOAT"
        return super().to_db_decimal_type(column)


class snowflake(Destination[SnowflakeClientConfiguration, "SnowflakeClient"]):
    spec = SnowflakeClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "jsonl"
        caps.supported_loader_file_formats = ["jsonl", "parquet", "csv", "model"]
        caps.preferred_staging_file_format = "jsonl"
        caps.supported_staging_file_formats = ["jsonl", "parquet", "csv"]
        caps.type_mapper = SnowflakeTypeMapper
        # snowflake is case sensitive but all unquoted identifiers are upper cased
        # so upper case identifiers are considered case insensitive
        caps.escape_identifier = escape_snowflake_identifier
        # dlt is configured to create case insensitive identifiers
        # note that case sensitive naming conventions will change this setting to "str" (case sensitive)
        caps.casefold_identifier = str.upper
        caps.has_case_sensitive_identifiers = True
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
        caps.max_identifier_length = 255
        caps.max_column_identifier_length = 255
        caps.max_query_length = 2 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 16 * 1024 * 1024
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = True
        caps.alter_add_multi_column = True
        caps.supports_clone_table = True
        caps.supported_merge_strategies = ["delete-insert", "upsert", "scd2"]
        caps.supported_replace_strategies = [
            "truncate-and-insert",
            "insert-from-staging",
            "staging-optimized",
        ]
        caps.timestamp_precision = 6
        caps.max_timestamp_precision = 9
        caps.sqlglot_dialect = "snowflake"

        return caps

    @property
    def client_class(self) -> Type["SnowflakeClient"]:
        from dlt.destinations.impl.snowflake.snowflake import SnowflakeClient

        return SnowflakeClient

    def __init__(
        self,
        credentials: Union[SnowflakeCredentials, Dict[str, Any], str] = None,
        stage_name: Optional[str] = None,
        keep_staged_files: bool = True,
        csv_format: Optional[CsvFormatConfiguration] = None,
        query_tag: Optional[str] = None,
        create_indexes: bool = False,
        use_decfloat: bool = False,
        enable_atomic_swap: bool = False,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Snowflake destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[SnowflakeCredentials, Dict[str, Any], str], optional): Credentials to connect to the snowflake database. Can be an instance of `SnowflakeCredentials` or
                a connection string in the format `snowflake://user:password@host:port/database`
            stage_name (Optional[str], optional): Name of an existing stage to use for loading data. Default uses implicit stage per table
            keep_staged_files (bool, optional): Whether to delete or keep staged files after loading
            csv_format (Optional[CsvFormatConfiguration]): Optional csv format configuration
            query_tag (Optional[str]): A tag with placeholders to tag sessions executing jobs
            create_indexes (bool, optional): Whether UNIQUE or PRIMARY KEY constrains should be created
            use_decfloat (bool, optional): Whether to use DECFLOAT type for unbound decimals. DECFLOAT stores
                exact decimal values with up to 36 significant digits and a dynamic exponent.
                Only works with text-based staging formats (jsonl, csv) - not parquet.
            enable_atomic_swap (bool, optional): Whether to use atomic swap when replacing with replace strategy `staging-optimized`.
            destination_name (str, optional): Name of the destination. Defaults to None.
            environment (str, optional): Environment name. Defaults to None.
            **kwargs (Any, optional): Additional arguments forwarded to the destination config
        """
        super().__init__(
            credentials=credentials,
            stage_name=stage_name,
            keep_staged_files=keep_staged_files,
            csv_format=csv_format,
            query_tag=query_tag,
            create_indexes=create_indexes,
            use_decfloat=use_decfloat,
            enable_atomic_swap=enable_atomic_swap,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )


snowflake.register()



from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    NewType,
    Union,
)
from typing_extensions import Never, NotRequired, Required

from dlt.common.data_types import TDataType
from dlt.common.normalizers.typing import TNormalizersConfig
from dlt.common.typing import (
    TSortOrder,
    TAnyDateTime,
    TLoaderFileFormat,
    TColumnNames,
    TypedDict,
    get_args,
)

try:
    from pydantic import BaseModel as _PydanticBaseModel
except ImportError:
    _PydanticBaseModel = Never  # type: ignore[assignment, misc]


# current version of schema engine
SCHEMA_ENGINE_VERSION = 11

# dlt tables
VERSION_TABLE_NAME = "_dlt_version"
LOADS_TABLE_NAME = "_dlt_loads"
PIPELINE_STATE_TABLE_NAME = "_dlt_pipeline_state"
DLT_NAME_PREFIX = "_dlt"

# default dlt columns
C_DLT_ID = "_dlt_id"
"""unique id of current row"""
C_DLT_LOAD_ID = "_dlt_load_id"
"""load id to identify records loaded in a single load package"""
# NOTE C_DLT_LOAD_ID != C_DLT_LOADS_TABLE_LOAD_ID but they refer to the exact same entity / value.
# They differ for backwards compatiblity reasons
# TODO add schema migration to use `_dlt_load_id` in `_dlt_loads` table
C_DLT_LOADS_TABLE_LOAD_ID = "load_id"
"""load id column in the table {LOADS_TABLE_NAME}. Meant to be joined with {C_DLT_LOAD_ID} of data tables"""
C_CHILD_PARENT_REF_LABEL = "_dlt_parent"
"""Label of the implicit `TTableReference` between a child table and its parent table"""
C_DESCENDANT_ROOT_REF_LABEL = "_dlt_root"
"""Label of the implicit `TTableReference` between a descendant table and its root table"""
C_ROOT_LOAD_REF_LABEL = "_dlt_load"
"""Label of the implicit `TTableReference` between a root table and the _dlt_loads table"""
C_VERSION_SCHEMA_VERSION_LABEL = "_dlt_schema_version"
"""Label of the implicit TTableReference between {LOAD_TABLE_NAME} and {VERSION_TABLE_NAME} for schema version."""
C_VERSION_SCHEMA_NAME_LABEL = "_dlt_schema_name"
"""Label of the implicit TTableReference between {LOAD_TABLE_NAME} and {VERSION_TABLE_NAME} for schema name."""

TColumnProp = Literal[
    "name",
    # data type
    "data_type",
    "precision",
    "scale",
    "timezone",
    "nullable",
    "variant",
    # hints
    "partition",
    "cluster",
    "primary_key",
    "sort",
    "unique",
    "merge_key",
    "row_key",
    "parent_key",
    "root_key",
    "hard_delete",
    "dedup_sort",
]
"""All known properties of the column, including name, data type info and hints"""
COLUMN_PROPS: Set[TColumnProp] = set(get_args(TColumnProp))

TColumnHint = Literal[
    "nullable",
    "partition",
    "cluster",
    "primary_key",
    "sort",
    "unique",
    "merge_key",
    "row_key",
    "parent_key",
    "root_key",
    "hard_delete",
    "dedup_sort",
]
"""Known hints of a column"""
COLUMN_HINTS: Set[TColumnHint] = set(get_args(TColumnHint))


class TColumnPropInfo(NamedTuple):
    name: Union[TColumnProp, str]
    defaults: Tuple[Any, ...] = (None,)
    is_hint: bool = False
    compound: bool = False


_ColumnPropInfos = [
    TColumnPropInfo("name"),
    TColumnPropInfo("data_type"),
    TColumnPropInfo("precision"),
    TColumnPropInfo("scale"),
    TColumnPropInfo("timezone", (True, None)),
    TColumnPropInfo("nullable", (True, None)),
    TColumnPropInfo("variant", (False, None)),
    TColumnPropInfo("partition", (False, None), False, True),
    TColumnPropInfo("cluster", (False, None), False, True),
    TColumnPropInfo("primary_key", (False, None), False, True),
    TColumnPropInfo("sort", (False, None)),
    TColumnPropInfo("unique", (False, None)),
    TColumnPropInfo("merge_key", (False, None), False, True),
    TColumnPropInfo("row_key", (False, None)),
    TColumnPropInfo("parent_key", (False, None)),
    TColumnPropInfo("root_key", (False, None)),
    TColumnPropInfo("hard_delete", (False, None)),
    TColumnPropInfo("dedup_sort", (False, None)),
    TColumnPropInfo("incremental", (False, None), False, True),
    # any x- hint with special settings ie. defaults
    TColumnPropInfo("x-active-record-timestamp", (), is_hint=True),  # no default values
]

ColumnPropInfos: Dict[Union[TColumnProp, str], TColumnPropInfo] = {
    info.name: info for info in _ColumnPropInfos
}
# verify column props and column hints infos
for hint in COLUMN_HINTS:
    assert hint in COLUMN_PROPS, f"Hint {hint} must be a column prop"

for prop in COLUMN_PROPS:
    assert prop in ColumnPropInfos, f"Column {prop} has no info, please define"
    if prop in COLUMN_HINTS:
        ColumnPropInfos[prop] = ColumnPropInfos[prop]._replace(is_hint=True)

TTableFormat = Literal["iceberg", "delta", "hive", "native"]
TFileFormat = Literal[Literal["preferred"], TLoaderFileFormat]
TTypeDetections = Literal[
    "timestamp", "iso_timestamp", "iso_date", "large_integer", "hexbytes_to_text", "wei_to_double"
]
TTypeDetectionFunc = Callable[[Type[Any], Any], Optional[TDataType]]


class TColumnType(TypedDict, total=False):
    data_type: Optional[TDataType]
    nullable: Optional[bool]
    precision: Optional[int]
    scale: Optional[int]
    timezone: Optional[bool]


# Part of Column containing processing hints added by the normalize stage
TColumnProcessingHints = TypedDict(
    "TColumnProcessingHints",
    {
        "x-normalizer": Optional[Dict[str, Any]],
    },
    total=False,
)


class TColumnSchemaBase(TColumnType, TColumnProcessingHints, total=False):
    """TypedDict that defines basic properties of a column: name, data type and nullable"""

    name: Optional[str]


class TColumnSchema(TColumnSchemaBase, total=False):
    """TypedDict that defines additional column hints"""

    description: Optional[str]
    partition: Optional[bool]
    cluster: Optional[bool]
    unique: Optional[bool]
    sort: Optional[bool]
    primary_key: Optional[bool]
    row_key: Optional[bool]
    parent_key: Optional[bool]
    root_key: Optional[bool]
    merge_key: Optional[bool]
    variant: Optional[bool]
    hard_delete: Optional[bool]
    dedup_sort: Optional[TSortOrder]
    incremental: Optional[bool]


TTableSchemaColumns = Dict[str, TColumnSchema]
"""A mapping from column name to column schema, typically part of a table schema"""


TAnySchemaColumns = Union[
    TTableSchemaColumns, Sequence[TColumnSchema], _PydanticBaseModel, Type[_PydanticBaseModel]
]

TSimpleRegex = NewType("TSimpleRegex", str)
TColumnName = NewType("TColumnName", str)
SIMPLE_REGEX_PREFIX = "re:"

TSchemaEvolutionMode = Literal["evolve", "discard_value", "freeze", "discard_row"]
TSchemaContractEntities = Literal["tables", "columns", "data_type"]


class TSchemaContractDict(TypedDict, total=False):
    """TypedDict defining the schema update settings"""

    tables: Optional[TSchemaEvolutionMode]
    columns: Optional[TSchemaEvolutionMode]
    data_type: Optional[TSchemaEvolutionMode]


TSchemaContract = Union[TSchemaEvolutionMode, TSchemaContractDict]


class TRowFilters(TypedDict, total=True):
    excludes: Optional[List[TSimpleRegex]]
    includes: Optional[List[TSimpleRegex]]


class NormalizerInfo(TypedDict, total=True):
    new_table: bool


# Part of Table containing processing hints added by pipeline stages
TTableProcessingHints = TypedDict(
    "TTableProcessingHints",
    {
        "x-normalizer": Optional[Dict[str, Any]],
        "x-loader": Optional[Dict[str, Any]],
        "x-extractor": Optional[Dict[str, Any]],
    },
    total=False,
)


TWriteDisposition = Literal["skip", "append", "replace", "merge"]
TLoaderMergeStrategy = Literal["delete-insert", "scd2", "upsert"]
TLoaderReplaceStrategy = Literal["truncate-and-insert", "insert-from-staging", "staging-optimized"]


WRITE_DISPOSITIONS: Sequence[TWriteDisposition] = sorted(get_args(TWriteDisposition))
MERGE_STRATEGIES: Sequence[TLoaderMergeStrategy] = sorted(get_args(TLoaderMergeStrategy))
REPLACE_STRATEGIES: Sequence[TLoaderReplaceStrategy] = sorted(get_args(TLoaderReplaceStrategy))

DEFAULT_VALIDITY_COLUMN_NAMES = ["_dlt_valid_from", "_dlt_valid_to"]
"""Default values for validity column names used in `scd2` merge strategy."""


class TWriteDispositionDict(TypedDict):
    disposition: TWriteDisposition


class TMergeDispositionDict(TWriteDispositionDict):
    strategy: Optional[TLoaderMergeStrategy]


class TDeleteInsertStrategyDict(TMergeDispositionDict):
    deduplicated: Optional[bool]


class TScd2StrategyDict(TMergeDispositionDict, total=False):
    validity_column_names: Optional[List[str]]
    active_record_timestamp: Optional[TAnyDateTime]
    boundary_timestamp: Optional[TAnyDateTime]
    row_version_column_name: Optional[str]


TWriteDispositionConfig = Union[
    TWriteDisposition,
    TWriteDispositionDict,
    TMergeDispositionDict,
    TScd2StrategyDict,
    TDeleteInsertStrategyDict,
]


TReferenceCardinality = Literal[
    "zero_to_one",
    "one_to_zero",
    "zero_to_many",
    "many_to_zero",
    "one_to_many",
    "many_to_one",
    "one_to_one",
    "many_to_many",
]
"""Represents cardinality between `column` (left) and `referenced_column` (right)

Note that cardinality is not symmetric. For example:
- `Author, 0 to many, Book` an author can have 0 to many book
- `Book, 1 to 1, Author` a book must have exactly 1 author

The statement (Author, 0 to many, Book) doesn't imply (Book, many to 0, Author).
"""


class _TTableReferenceBase(TypedDict, total=False):
    """Describes a reference to another table's columns.
    `columns` corresponds to the `referenced_columns` in the referenced table and their order should match.
    """

    label: Optional[str]
    """Text providing semantic information about the reference.

    For example, the label "liked" describe the relationship between `user` and `post` (user.id, "liked", post.id)
    """

    cardinality: Optional[TReferenceCardinality]
    """Cardinality of the relationship between `table.column` (left) and `referenced_table.referenced_column` (right)."""

    columns: Sequence[str]
    """Name of the column(s) from `table`"""

    referenced_table: str
    """Name of the referenced table"""

    referenced_columns: Sequence[str]
    """Name of the columns(s) from `referenced_table`"""


class TTableReferenceInline(_TTableReferenceBase, TypedDict, total=False):
    table: Optional[str]
    """Name of the table.
    When `TTableReference` is defined on a `TTableSchema` (i.e., "inline reference"), the `table`
    value is determined by `TTableSchema["name"]`
    """


# Keep for backwards compatibility
TTableReference = TTableReferenceInline


# Compared to `TTableReference` or `TInlineTableReference`, `table` is required
class TTableReferenceStandalone(_TTableReferenceBase, TypedDict, total=False):
    table: str
    """Name of the table.
    When `TTableReference` is defined on a `TTableSchema` (i.e., "inline reference"), the `table`
    value is determined by `TTableSchema["name"]`
    """


TTableReferenceParam = Sequence[TTableReference]


class _TTableSchemaBase(TTableProcessingHints, total=False):
    name: Optional[str]
    description: Optional[str]
    schema_contract: Optional[TSchemaContract]
    table_sealed: Optional[bool]
    parent: Optional[str]
    filters: Optional[TRowFilters]
    columns: TTableSchemaColumns
    resource: Optional[str]
    table_format: Optional[TTableFormat]
    file_format: Optional[TFileFormat]


class TTableSchema(_TTableSchemaBase, total=False):
    """TypedDict that defines properties of a table"""

    write_disposition: Optional[TWriteDisposition]
    references: Optional[TTableReferenceParam]


class TPartialTableSchema(TTableSchema):
    pass


TSchemaTables = Dict[str, TTableSchema]
TSchemaUpdate = Dict[str, List[TPartialTableSchema]]
TColumnDefaultHint = Literal["not_null", TColumnHint]
"""Allows using not_null in default hints setting section"""


class TSchemaSettings(TypedDict, total=False):
    schema_contract: Optional[TSchemaContract]
    detections: Optional[List[TTypeDetections]]
    default_hints: Optional[Dict[TColumnDefaultHint, List[TSimpleRegex]]]
    preferred_types: Optional[Dict[TSimpleRegex, TDataType]]


class TStoredSchema(TypedDict, total=False):
    """TypeDict defining the schema representation in storage"""

    version: int
    version_hash: str
    previous_hashes: List[str]
    imported_version_hash: Optional[str]
    engine_version: int
    name: str
    description: Optional[str]
    settings: Optional[TSchemaSettings]
    tables: TSchemaTables
    normalizers: TNormalizersConfig



from packaging.version import Version
from collections.abc import Mapping
from typing import Optional, Dict, Union, List
from pathlib import Path

from dlt import version, Pipeline
from dlt.common import logger
from dlt.common.libs.pyarrow import pyarrow as pa
from dlt.common.libs.pyarrow import cast_arrow_schema_types
from dlt.common.libs.utils import load_open_tables
from dlt.common.schema.typing import TWriteDisposition, TTableSchema
from dlt.common.schema.utils import get_first_column_name_with_prop, get_columns_names_with_prop
from dlt.common.exceptions import MissingDependencyException, ValueErrorWithKnownValues
from dlt.common.typing import DictStrAny
from dlt.common.utils import assert_min_pkg_version
from dlt.common.configuration.specs import CredentialsConfiguration
from dlt.common.configuration.specs.mixins import WithObjectStoreRsCredentials

try:
    import deltalake
    from deltalake import write_deltalake, DeltaTable

    deltalake_semver = Version(deltalake.__version__)
except ModuleNotFoundError:
    raise MissingDependencyException(
        "dlt deltalake helpers",
        [f"{version.DLT_PKG_NAME}[deltalake]"],
        "Install `deltalake` so dlt can create Delta tables in the `filesystem` destination.",
    )


def ensure_delta_compatible_arrow_schema(
    schema: pa.Schema,
    partition_by: Optional[Union[List[str], str]] = None,
) -> pa.Schema:
    """Returns Arrow schema compatible with Delta table format.

    Casts schema to replace data types not supported by Delta.
    """
    ARROW_TO_DELTA_COMPATIBLE_ARROW_TYPE_MAP = {
        # maps type check function to type factory function
        pa.types.is_null: pa.string(),
        pa.types.is_time: pa.string(),
        pa.types.is_decimal256: pa.string(),  # pyarrow does not allow downcasting to decimal128
    }

    # partition fields can't be dictionary: https://github.com/delta-io/delta-rs/issues/2969
    if partition_by is not None:
        if isinstance(partition_by, str):
            partition_by = [partition_by]
        if any(pa.types.is_dictionary(schema.field(col).type) for col in partition_by):
            # cast all dictionary fields to string — this is rogue because
            # 1. dictionary value type is disregarded
            # 2. any non-partition dictionary fields are cast too
            ARROW_TO_DELTA_COMPATIBLE_ARROW_TYPE_MAP[pa.types.is_dictionary] = (
                lambda t_: t_.value_type
            )

    # NOTE: also consider calling _convert_pa_schema_to_delta() from delta.schema which casts unsigned types
    return cast_arrow_schema_types(schema, ARROW_TO_DELTA_COMPATIBLE_ARROW_TYPE_MAP)


def ensure_delta_compatible_arrow_data(
    data: Union[pa.Table, pa.RecordBatchReader],
    partition_by: Optional[Union[List[str], str]] = None,
) -> Union[pa.Table, pa.RecordBatchReader]:
    """Returns Arrow data compatible with Delta table format.

    Casts `data` schema to replace data types not supported by Delta.
    """
    # RecordBatchReader.cast() requires pyarrow>=17.0.0
    # cast() got introduced in 16.0.0, but with bug
    assert_min_pkg_version(
        pkg_name="pyarrow",
        version="17.0.0",
        msg="`pyarrow>=17.0.0` is needed for `delta` table format on `filesystem` destination.",
    )
    schema = ensure_delta_compatible_arrow_schema(data.schema, partition_by)
    return data.cast(schema)


def get_delta_write_mode(write_disposition: TWriteDisposition) -> str:
    """Translates dlt write disposition to Delta write mode."""
    if write_disposition in ("append", "merge"):  # `merge` disposition resolves to `append`
        return "append"
    elif write_disposition == "replace":
        return "overwrite"
    else:
        raise ValueErrorWithKnownValues(
            "write_disposition", write_disposition, ["append", "replace", "merge"]
        )


def write_delta_table(
    table_or_uri: Union[str, Path, DeltaTable],
    data: Union[pa.Table, pa.RecordBatchReader],
    write_disposition: TWriteDisposition,
    partition_by: Optional[Union[List[str], str]] = None,
    storage_options: Optional[Dict[str, str]] = None,
    configuration: Optional[Mapping[str, Optional[str]]] = None,
) -> None:
    """Writes in-memory Arrow data to on-disk Delta table.

    Thin wrapper around `deltalake.write_deltalake`.
    """
    write_deltalake(  # type: ignore[call-overload]
        table_or_uri=table_or_uri,
        data=ensure_delta_compatible_arrow_data(data, partition_by),
        partition_by=partition_by,
        mode=get_delta_write_mode(write_disposition),
        schema_mode="merge",  # enable schema evolution (adding new columns)
        storage_options=storage_options,
        configuration=configuration,
    )


def merge_delta_table(
    table: DeltaTable,
    data: Union[pa.Table, pa.RecordBatchReader],
    schema: TTableSchema,
    load_table_name: str,
    streamed_exec: bool,
) -> None:
    """Merges in-memory Arrow data into on-disk Delta table."""

    strategy = schema["x-merge-strategy"]  # type: ignore[typeddict-item]
    if strategy == "upsert":
        # `DeltaTable.merge` does not support automatic schema evolution
        # https://github.com/delta-io/delta-rs/issues/2282
        # NOTE: fixing the issue didn't help here
        evolve_delta_table_schema(table, data.schema)

        if "parent" in schema:
            unique_column = get_first_column_name_with_prop(schema, "unique")
            predicate = f"target.{unique_column} = source.{unique_column}"
        else:
            primary_keys = get_columns_names_with_prop(schema, "primary_key")
            predicate = " AND ".join([f"target.{c} = source.{c}" for c in primary_keys])

        partition_by = get_columns_names_with_prop(schema, "partition")
        qry = (
            table.merge(
                source=ensure_delta_compatible_arrow_data(data, partition_by),
                predicate=predicate,
                source_alias="source",
                target_alias="target",
                streamed_exec=streamed_exec,
            )
            .when_matched_update_all()
            .when_not_matched_insert_all()
        )

        qry.execute()
    else:
        raise ValueError(
            f'Merge strategy "{strategy}" is not supported for Delta tables. '
            f'Table: "{load_table_name}".'
        )


def get_delta_tables(
    pipeline: Pipeline, *tables: str, schema_name: str = None, include_dlt_tables: bool = False
) -> Dict[str, DeltaTable]:
    """Returns Delta tables in `pipeline.default_schema (default)` or `schema_name` as `deltalake.DeltaTable` objects.

    Returned object is a dictionary with table names as keys and `DeltaTable` objects as values.
    Optionally filters dictionary by table names specified as `*tables*`.
    Raises ValueError if table name specified as `*tables` is not found. You may try to switch to other
    schemas via `schema_name` argument.
    """
    return load_open_tables(
        pipeline, "delta", *tables, schema_name=schema_name, include_dlt_tables=include_dlt_tables
    )


def deltalake_storage_options(
    credentials: Optional[CredentialsConfiguration] = None,
    storage_options: Optional[DictStrAny] = None,
) -> Dict[str, str]:
    """Returns dict that can be passed as `storage_options` in `deltalake` library."""
    creds = {}
    extra_options = {}
    if isinstance(credentials, WithObjectStoreRsCredentials):
        creds = credentials.to_object_store_rs_credentials()
    if storage_options is not None:
        extra_options = storage_options
    shared_keys = creds.keys() & extra_options.keys()
    if len(shared_keys) > 0:
        logger.warning(
            "The `deltalake_storage_options` configuration dictionary contains "
            "keys also provided by dlt's credential system: "
            + ", ".join([f"`{key}`" for key in shared_keys])
            + ". dlt will use the values in `deltalake_storage_options`."
        )
    return {**creds, **extra_options}


def evolve_delta_table_schema(delta_table: DeltaTable, arrow_schema: pa.Schema) -> DeltaTable:
    """Evolves `delta_table` schema if different from `arrow_schema`.

    We compare fields via names. Actual types and nullability are ignored. This is
    how schemas are evolved for other destinations. Existing columns are never modified.
    Variant columns are created.

    Adds column(s) to `delta_table` present in `arrow_schema` but not in `delta_table`.
    """

    if deltalake_semver.major == 0:
        new_fields = [
            deltalake.Field.from_pyarrow(field)  # type: ignore[attr-defined]
            for field in ensure_delta_compatible_arrow_schema(arrow_schema)
            if field.name not in delta_table.schema().to_pyarrow().names  # type: ignore[attr-defined]
        ]
    else:
        # deltalake 1.x changed pyarrow to arrow
        new_fields = [
            deltalake.Field.from_arrow(field)
            for field in ensure_delta_compatible_arrow_schema(arrow_schema)
            if field.name not in delta_table.schema().to_arrow().names
        ]
    if new_fields:
        delta_table.alter.add_columns(new_fields)
    return delta_table



from datetime import datetime, timezone, timedelta, tzinfo  # noqa: I251
from typing import Optional, SupportsIndex, Union
import pendulum  # noqa: I251
from pendulum.tz import UTC
from pendulum.tz.timezone import FixedTimezone, Timezone

# force UTC as the local timezone to prevent local dates to be written to dbs
pendulum.set_local_timezone(pendulum.timezone("UTC"))


def __utcnow() -> pendulum.DateTime:
    """
    Use this function instead of datetime.now
    Returns:
        pendulum.DateTime -- current time in UTC timezone
    """
    return pendulum.now()


def to_pendulum_tz(tz: Optional[tzinfo]) -> Optional[Union[Timezone, FixedTimezone]]:
    """Convert a tzinfo to a pendulum Timezone/FixedTimezone"""
    if tz is None:
        return None
    # already a pendulum timezone
    if isinstance(tz, (Timezone, FixedTimezone)):
        return tz
    # Python stdlib fixed offset - use cached fixed_timezone
    if isinstance(tz, timezone):
        offset_seconds = int(tz.utcoffset(None).total_seconds())
        if offset_seconds == 0:
            return UTC
        return pendulum.fixed_timezone(offset_seconds)
    # named timezone (pytz, dateutil, zoneinfo) - need _safe_timezone for DST
    return pendulum._safe_timezone(tz)


def create_dt(
    year: SupportsIndex,
    month: SupportsIndex,
    day: SupportsIndex,
    hour: SupportsIndex = 0,
    minute: SupportsIndex = 0,
    second: SupportsIndex = 0,
    microsecond: SupportsIndex = 0,
    tz: tzinfo = None,
    fold: int = 1,
) -> pendulum.DateTime:
    """Creates a new DateTime instance from a specific date and time."""
    pend_tz = to_pendulum_tz(tz)

    if pend_tz is None:
        # naive datetime
        return pendulum.DateTime(year, month, day, hour, minute, second, microsecond)

    if isinstance(pend_tz, FixedTimezone) or pend_tz is UTC:
        # fixed offset or UTC - no DST, no conversion needed
        return pendulum.DateTime(
            year, month, day, hour, minute, second, microsecond, tzinfo=pend_tz
        )

    # named timezone - need convert() for DST handling
    dt = datetime(year, month, day, hour, minute, second, microsecond, fold=fold)
    dt = pend_tz.convert(dt)
    return pendulum.DateTime(
        dt.year,
        dt.month,
        dt.day,
        dt.hour,
        dt.minute,
        dt.second,
        dt.microsecond,
        tzinfo=dt.tzinfo,
        fold=dt.fold,
    )


def ensure_pendulum_dt(dt: datetime) -> pendulum.DateTime:
    """Convert a stdlib datetime to pendulum DateTime preserving timezone."""
    tz = dt.tzinfo
    if tz is None:
        return pendulum.DateTime(
            dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond
        )
    if tz is UTC:
        return pendulum.DateTime(
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            dt.microsecond,
            tzinfo=UTC,
        )
    return create_dt(
        dt.year,
        dt.month,
        dt.day,
        dt.hour,
        dt.minute,
        dt.second,
        dt.microsecond,
        tz=tz,
        fold=dt.fold,
    )


pendulum.utcnow = __utcnow  # type: ignore



from abc import ABC, abstractmethod
from contextlib import contextmanager
from functools import wraps
import inspect
from types import TracebackType
from typing import (
    Any,
    ClassVar,
    ContextManager,
    Dict,
    Generic,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Type,
    AnyStr,
    List,
    Generator,
    cast,
)

from dlt.common.destination.exceptions import DestinationUndefinedEntity
from dlt.common.typing import TFun, TypedDict, Self
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.utils import concat_strings_with_limit
from dlt.common.destination.client import JobClientBase

from dlt.destinations.exceptions import (
    DestinationConnectionError,
    LoadClientNotConnected,
)
from dlt.destinations.typing import (
    DBApi,
    TNativeConn,
    DataFrame,
    DBTransaction,
    ArrowTable,
)
from dlt.common.destination.dataset import DBApiCursor


class TJobQueryTags(TypedDict):
    """Applied to sql client when a job using it starts. Using to tag queries"""

    source: str
    resource: str
    table: str
    load_id: str
    pipeline_name: str


class SqlClientBase(ABC, Generic[TNativeConn]):
    dbapi: ClassVar[DBApi] = None

    database_name: Optional[str]
    """Database or catalog name, optional"""
    dataset_name: str
    """Normalized dataset name"""
    staging_dataset_name: str
    """Normalized staging dataset name"""
    capabilities: DestinationCapabilitiesContext
    """Instance of adjusted destination capabilities"""

    def __init__(
        self,
        database_name: str,
        dataset_name: str,
        staging_dataset_name: str,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        self.dataset_name = dataset_name
        self.staging_dataset_name = staging_dataset_name
        self.database_name = database_name
        self.capabilities = capabilities
        self._query_tags: TJobQueryTags = None

    @abstractmethod
    def open_connection(self) -> TNativeConn:
        pass

    @abstractmethod
    def close_connection(self) -> None:
        pass

    @abstractmethod
    def begin_transaction(self) -> ContextManager[DBTransaction]:
        pass

    def __getattr__(self, name: str) -> Any:
        # pass unresolved attrs to native connections
        if not self.native_connection:
            raise AttributeError(name)
        return getattr(self.native_connection, name)

    def __enter__(self) -> Self:
        self.open_connection()
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        self.close_connection()

    @property
    @abstractmethod
    def native_connection(self) -> TNativeConn:
        pass

    def has_dataset(self) -> bool:
        query = f"""
SELECT 1
    FROM {self._qualify_info_schema_table_name("SCHEMATA")}
    WHERE """
        catalog_name, schema_name, _ = self._get_information_schema_components()
        db_params: List[str] = []
        if catalog_name is not None:
            query += " catalog_name = %s AND "
            db_params.append(catalog_name)
        db_params.append(schema_name)
        query += "schema_name = %s"
        rows = self.execute_sql(query, *db_params)
        return len(rows) > 0

    def create_dataset(self) -> None:
        self.execute_sql("CREATE SCHEMA %s" % self.fully_qualified_dataset_name())

    def drop_dataset(self) -> None:
        # assert self.fully_qualified_dataset_name() != "None"
        self.execute_sql("DROP SCHEMA %s CASCADE" % self.fully_qualified_dataset_name())

    def truncate_tables(self, *tables: str) -> None:
        statements = [self._truncate_table_sql(self.make_qualified_table_name(t)) for t in tables]
        self.execute_many(statements)

    def drop_tables(self, *tables: str) -> None:
        """Drops a set of tables if they exist"""
        if not tables:
            return
        statements = [
            f"DROP TABLE IF EXISTS {self.make_qualified_table_name(table)}" for table in tables
        ]
        self.execute_many(statements)

    def _to_named_paramstyle(self, query: str, args: Sequence[Any]) -> Tuple[str, Dict[str, Any]]:
        """Convert a query from "format" ( %s ) paramstyle to "named" ( :param_name ) paramstyle.
        The %s are replaced with :arg0, :arg1, ... and the arguments are returned as a dictionary.

        Args:
            query: SQL query with %s placeholders
            args: arguments to be passed to the query

        Returns:
            Tuple of the new query and a dictionary of named arguments
        """
        keys = [f"arg{i}" for i in range(len(args))]
        # Replace position arguments (%s) with named arguments (:arg0, :arg1, ...)
        query = query % tuple(f":{key}" for key in keys)
        db_args = {key: db_arg for key, db_arg in zip(keys, args)}
        return query, db_args

    @abstractmethod
    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        pass

    @abstractmethod
    def execute_query(
        self, query: AnyStr, *args: Any, **kwargs: Any
    ) -> ContextManager[DBApiCursor]:
        pass

    def execute_fragments(
        self, fragments: Sequence[AnyStr], *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        """Executes several SQL fragments as efficiently as possible to prevent data copying. Default implementation just joins the strings and executes them together."""
        return self.execute_sql("".join(fragments), *args, **kwargs)  # type: ignore

    def execute_many(
        self, statements: Sequence[str], *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        """Executes multiple SQL statements as efficiently as possible. When client supports multiple statements in a single query
        they are executed together in as few database calls as possible.
        """
        ret = []
        if self.capabilities.supports_multiple_statements:
            for sql_fragment in concat_strings_with_limit(
                list(statements), ";\n", self.capabilities.max_query_length // 2
            ):
                ret.append(self.execute_sql(sql_fragment + ";", *args, **kwargs))
        else:
            for statement in statements:
                result = self.execute_sql(statement, *args, **kwargs)
                if result is not None:
                    ret.append(result)
        return ret

    # TODO make it a staticmethod to avoid passing SQLClient instances all around
    def catalog_name(self, quote: bool = True, casefold: bool = True) -> Optional[str]:
        # default is no catalogue component of the name, which typically means that
        # connection is scoped to a current database
        return None

    # TODO make it a staticmethod to avoid passing SQLClient instances all around
    def fully_qualified_dataset_name(
        self, quote: bool = True, staging: bool = False, casefold: bool = True
    ) -> str:
        if staging:
            with self.with_staging_dataset():
                path = self.make_qualified_table_name_path(None, quote=quote, casefold=casefold)
        else:
            path = self.make_qualified_table_name_path(None, quote=quote, casefold=casefold)
        return ".".join(path)

    # TODO make it a staticmethod to avoid passing SQLClient instances all around
    def make_qualified_table_name(
        self, table_name: str, quote: bool = True, casefold: bool = True
    ) -> str:
        return ".".join(
            self.make_qualified_table_name_path(table_name, quote=quote, casefold=casefold)
        )

    # TODO make it a staticmethod to avoid passing SQLClient instances all around
    def make_qualified_table_name_path(
        self, table_name: Optional[str], quote: bool = True, casefold: bool = True
    ) -> List[str]:
        """Returns a list with path components leading from catalog to table_name.
        Used to construct fully qualified names. `table_name` is optional.
        """
        path: List[str] = []
        if catalog_name := self.catalog_name(quote=quote, casefold=casefold):
            path.append(catalog_name)
        dataset_name = self.dataset_name
        if casefold:
            dataset_name = self.capabilities.casefold_identifier(self.dataset_name)
        if quote:
            dataset_name = self.capabilities.escape_identifier(dataset_name)
        path.append(dataset_name)
        if table_name:
            if casefold:
                table_name = self.capabilities.casefold_identifier(table_name)
            if quote:
                table_name = self.capabilities.escape_identifier(table_name)
            path.append(table_name)
        return path

    def get_qualified_table_names(
        self, table_name: str, quote: bool = True, casefold: bool = True
    ) -> Tuple[str, str]:
        """Returns qualified names for table and corresponding staging table as tuple."""
        with self.with_staging_dataset():
            staging_table_name = self.make_qualified_table_name(
                table_name, quote=quote, casefold=casefold
            )
        return (
            self.make_qualified_table_name(table_name, quote=quote, casefold=casefold),
            staging_table_name,
        )

    def escape_column_name(
        self, column_name: str, quote: bool = True, casefold: bool = True
    ) -> str:
        if casefold:
            column_name = self.capabilities.casefold_identifier(column_name)
        if quote:
            return self.capabilities.escape_identifier(column_name)
        return column_name

    @contextmanager
    def with_alternative_dataset_name(
        self, dataset_name: str
    ) -> Iterator["SqlClientBase[TNativeConn]"]:
        """Sets the `dataset_name` as the default dataset during the lifetime of the context. Does not modify any search paths in the existing connection."""
        current_dataset_name = self.dataset_name
        try:
            self.dataset_name = dataset_name
            yield self
        finally:
            # restore previous dataset name
            self.dataset_name = current_dataset_name

    def with_staging_dataset(self) -> ContextManager["SqlClientBase[TNativeConn]"]:
        """Temporarily switch sql client to staging dataset name"""
        return self.with_alternative_dataset_name(self.staging_dataset_name)

    @property
    def is_staging_dataset_active(self) -> bool:
        """Checks if staging dataset is currently active"""
        return self.dataset_name == self.staging_dataset_name

    def set_query_tags(self, tags: TJobQueryTags) -> None:
        """Sets current schema (source), resource, load_id and table name when a job starts"""
        self._query_tags = tags

    def _ensure_native_conn(self) -> None:
        if not self.native_connection:
            raise LoadClientNotConnected(type(self).__name__, self.dataset_name)

    @staticmethod
    @abstractmethod
    def _make_database_exception(ex: Exception) -> Exception:
        pass

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        # crude way to detect dbapi DatabaseError: there's no common set of exceptions, each module must reimplement
        mro = type.mro(type(ex))
        return any(t.__name__ in ("DatabaseError", "DataError") for t in mro)

    def _qualify_info_schema_table_name(self, table_name: str) -> str:
        return f"INFORMATION_SCHEMA.{table_name}"

    def _get_information_schema_components(self, *tables: str) -> Tuple[str, str, List[str]]:
        """Gets catalog name, schema name and name of the tables in format that can be directly
        used to query INFORMATION_SCHEMA. catalog name is optional: in that case None is
        returned in the first element of the tuple.
        """
        schema_path = self.make_qualified_table_name_path(None, quote=False)
        return (
            self.catalog_name(quote=False),
            schema_path[-1],
            [self.make_qualified_table_name_path(table, quote=False)[-1] for table in tables],
        )

    #
    # generate sql statements
    #
    def _truncate_table_sql(self, qualified_table_name: str) -> str:
        if self.capabilities.supports_truncate_command:
            return f"TRUNCATE TABLE {qualified_table_name}"
        else:
            return f"DELETE FROM {qualified_table_name} WHERE 1=1"

    def _limit_clause_sql(self, limit: int) -> Tuple[str, str]:
        return "", f"LIMIT {limit}"


class WithSqlClient(ABC):
    @property
    @abstractmethod
    def sql_client(self) -> SqlClientBase[TNativeConn]:
        pass

    # Sqlclient type so we know when WithTableScanners is implemented
    @property
    @abstractmethod
    def sql_client_class(self) -> Type[SqlClientBase[TNativeConn]]:
        pass


class DBApiCursorImpl(DBApiCursor):
    """A DBApi Cursor wrapper with dataframes reading functionality"""

    def __init__(self, curr: DBApiCursor) -> None:
        self.native_cursor = curr
        self._set_default_schema_columns()

    def __getattr__(self, name: str) -> Any:
        return getattr(self.native_cursor, name)

    def _get_columns(self) -> List[str]:
        if self.native_cursor.description:
            return [c[0] for c in self.native_cursor.description]
        return []

    # wire protocol methods
    def execute(self, *args: Any, **kwargs: Any) -> None:
        self.native_cursor.execute(*args, **kwargs)

    def fetchall(self, *args: Any, **kwargs: Any) -> List[Tuple[Any, ...]]:
        return self.native_cursor.fetchall(*args, **kwargs)

    def fetchmany(self, *args: Any, **kwargs: Any) -> List[Tuple[Any, ...]]:
        return self.native_cursor.fetchmany(*args, **kwargs)

    def fetchone(self, *args: Any, **kwargs: Any) -> Tuple[Any, ...]:
        return self.native_cursor.fetchone(*args, **kwargs)

    def close(self, *args: Any, **kwargs: Any) -> None:
        self.native_cursor.close(*args, **kwargs)

    def _set_default_schema_columns(self) -> None:
        self.columns_schema = cast(
            TTableSchemaColumns, {c: {"name": c, "nullable": True} for c in self._get_columns()}
        )

    def df(self, chunk_size: int = None, **kwargs: Any) -> Optional[DataFrame]:
        """Fetches results as data frame in full or in specified chunks.

        May use native pandas/arrow reader if available. Depending on
        the native implementation chunk size may vary.
        """
        try:
            return next(self.iter_df(chunk_size=chunk_size))
        except StopIteration:
            return None

    def arrow(self, chunk_size: int = None, **kwargs: Any) -> Optional[ArrowTable]:
        """Fetches results as data frame in full or in specified chunks.

        May use native pandas/arrow reader if available. Depending on
        the native implementation chunk size may vary.
        """
        try:
            return next(self.iter_arrow(chunk_size=chunk_size))
        except StopIteration:
            return None

    def iter_fetch(self, chunk_size: int) -> Generator[List[Tuple[Any, ...]], Any, Any]:
        while True:
            if not (result := self.fetchmany(chunk_size)):
                return
            yield result

    def iter_df(self, chunk_size: int) -> Generator[DataFrame, None, None]:
        """Default implementation converts arrow to df"""
        from dlt.common.libs.pandas import pandas as pd

        for table in self.iter_arrow(chunk_size=chunk_size):
            # NOTE: we go via arrow table, types are created for arrow is columns are known
            # https://github.com/apache/arrow/issues/38644 for reference on types_mapper
            yield table.to_pandas()

    def iter_arrow(self, chunk_size: int) -> Generator[ArrowTable, None, None]:
        """Default implementation converts query result to arrow table"""
        from dlt.common.libs.pyarrow import row_tuples_to_arrow
        from dlt.common.configuration.container import Container

        # get capabilities of possibly currently active pipeline
        caps = (
            Container().get(DestinationCapabilitiesContext)
            or DestinationCapabilitiesContext.generic_capabilities()
        )

        # note that columns_schema is inferred from the cursor and is not normalized
        if not chunk_size:
            result = self.fetchall()
            yield row_tuples_to_arrow(result, caps, self.columns_schema, tz="UTC")
            return

        for result in self.iter_fetch(chunk_size=chunk_size):
            yield row_tuples_to_arrow(result, caps, self.columns_schema, tz="UTC")


def raise_database_error(f: TFun) -> TFun:
    @wraps(f)
    def _wrap_gen(self: SqlClientBase[Any], *args: Any, **kwargs: Any) -> Any:
        try:
            return (yield from f(self, *args, **kwargs))
        except Exception as ex:
            db_ex = self._make_database_exception(ex)
            if db_ex is ex:
                raise db_ex.with_traceback(ex.__traceback__)
            else:
                raise db_ex.with_traceback(ex.__traceback__) from ex

    @wraps(f)
    def _wrap(self: SqlClientBase[Any], *args: Any, **kwargs: Any) -> Any:
        try:
            return f(self, *args, **kwargs)
        except Exception as ex:
            db_ex = self._make_database_exception(ex)
            if db_ex is ex:
                raise db_ex.with_traceback(ex.__traceback__)
            else:
                raise db_ex.with_traceback(ex.__traceback__) from ex

    if inspect.isgeneratorfunction(f):
        return _wrap_gen  # type: ignore[return-value]
    return _wrap  # type: ignore[return-value]


def raise_open_connection_error(f: TFun) -> TFun:
    @wraps(f)
    def _wrap(self: SqlClientBase[Any], *args: Any, **kwargs: Any) -> Any:
        try:
            return f(self, *args, **kwargs)
        except Exception as ex:
            db_ex = self._make_database_exception(ex)
            if isinstance(db_ex, DestinationUndefinedEntity):
                if db_ex is ex:
                    raise db_ex.with_traceback(ex.__traceback__)
                else:
                    raise db_ex.with_traceback(ex.__traceback__) from ex
            raise DestinationConnectionError(type(self).__name__, self.dataset_name, str(ex), ex)

    return _wrap  # type: ignore



from typing import Dict, Any
from abc import abstractmethod, ABC

from dlt.common.typing import Self


class WithObjectStoreRsCredentials(ABC):
    @abstractmethod
    def to_object_store_rs_credentials(self) -> Dict[str, Any]:
        """Returns credentials dictionary for object_store Rust crate.

        Can be used for libraries that build on top of the object_store crate, such as `deltalake`.

        https://docs.rs/object_store/latest/object_store/
        """


class WithPyicebergConfig(ABC):
    @abstractmethod
    def to_pyiceberg_fileio_config(self) -> Dict[str, Any]:
        """Returns `pyiceberg` FileIO configuration dictionary.

        https://py.iceberg.apache.org/configuration/#fileio
        """

    @classmethod
    @abstractmethod
    def from_pyiceberg_fileio_config(cls, file_io: Dict[str, Any]) -> Self:
        """Returns credentials instance from FileIO. If `file_io` does
        not fit into given credential type, credentials remain unresolved.
        """



import re
from typing import ClassVar

from dlt.common.typing import REPattern
from dlt.common.normalizers.naming.naming import NamingConvention as BaseNamingConvention


RE_UNDERSCORES = re.compile("__+")
RE_LEADING_DIGITS = re.compile(r"^\d+")
RE_ENDING_UNDERSCORES = re.compile(r"_+$")
RE_NON_ALPHANUMERIC = re.compile(r"[^a-zA-Z\d_]+")


class NamingConvention(BaseNamingConvention):
    """Generates case sensitive SQL safe identifiers, preserving the source casing.

    - Spaces around identifier are trimmed
    - Removes all ascii characters except ascii alphanumerics and underscores
    - Prepends `_` if name starts with number.
    - Removes all trailing underscores.
    - Multiples of `_` are converted into single `_`.
    """

    RE_NON_ALPHANUMERIC: ClassVar[REPattern] = RE_NON_ALPHANUMERIC
    RE_UNDERSCORES: ClassVar[REPattern] = RE_UNDERSCORES
    RE_ENDING_UNDERSCORES: ClassVar[REPattern] = RE_ENDING_UNDERSCORES

    def normalize_identifier(self, identifier: str) -> str:
        identifier = super().normalize_identifier(identifier)
        # remove non alpha characters
        norm_identifier = self.RE_NON_ALPHANUMERIC.sub("_", identifier)
        # remove leading digits
        if RE_LEADING_DIGITS.match(norm_identifier):
            norm_identifier = "_" + norm_identifier
        # remove trailing underscores to not mess with how we break paths
        if norm_identifier != "_":
            norm_identifier = self.RE_ENDING_UNDERSCORES.sub("", norm_identifier)
        # contract multiple __
        norm_identifier = self.RE_UNDERSCORES.sub("_", norm_identifier)
        return self.shorten_identifier(norm_identifier, identifier, self.max_length)

    @property
    def is_case_sensitive(self) -> bool:
        return True



"""
JSON Column Expansion Module

Pure transformation layer for expanding JSON string columns into dictionaries
before they enter DLT's flattening pipeline.
"""

import json
from typing import Any, Dict, List, Optional, Tuple, Union

from dlt.common import logger
from dlt.common.typing import DictStrAny


def parse_json_value(value: Any) -> Optional[DictStrAny]:
    """Parse JSON string or dict-like value into a dict."""
    if value is None:
        return None
    if isinstance(value, dict):
        return value
    try:
        return json.loads(value)
    except Exception:
        logger.error(
            "Invalid JSON value for JSON expansion, returning original value without flattening"
        )
        return None


def filter_by_paths(data: Dict[str, Any], paths: List[str]) -> Dict[str, Any]:
    """Extract only specified paths from dict, preserving nested structure."""
    result: Dict[str, Any] = {}

    for path in paths:
        current = data
        parts = path.split(".")
        valid = True

        for p in parts:
            if not isinstance(current, dict) or p not in current:
                valid = False
                break
            current = current[p]

        if not valid:
            continue

        # rebuild nested structure
        d = result
        for p in parts[:-1]:
            d = d.setdefault(p, {})
        d[parts[-1]] = current

    return result


def expand_json_column(
    raw_value: Any,
    flatten_spec: Union[bool, List[str], None],
    keep_original: bool,
) -> Tuple[Any, Optional[Dict[str, Any]]]:
    """
    Expand JSON column according to flatten specification.

    Returns:
        Tuple of (original_value_to_store, dict_to_flatten)
    """
    if not flatten_spec:
        return raw_value, None

    parsed_value = parse_json_value(raw_value)

    if parsed_value is None:
        return raw_value, None

    if flatten_spec is True:
        return raw_value if keep_original else None, parsed_value

    if isinstance(flatten_spec, list):
        filtered = filter_by_paths(parsed_value, flatten_spec)
        return raw_value if keep_original else None, filtered

    return raw_value, None



from typing import List

from pydantic import Field

from dlt.common.typing import Annotated
from dlt._workspace.cli._ai_context_api_client import search_sources
from dlt._workspace.cli.exceptions import AiContextApiError
from dlt._workspace.mcp.context import with_mcp_tool_telemetry
from dlt._workspace.typing import TSourceItem

from fastmcp.exceptions import ToolError


@with_mcp_tool_telemetry()
def search_dlthub_sources(
    query: Annotated[
        str,
        Field(
            description=(
                "Search term to find sources by name or description. Empty string returns all"
                " sources alphabetically."
            )
        ),
    ] = "",
) -> List[TSourceItem]:
    """Search for available dlt sources on dlthub by name or description."""
    try:
        return search_sources(query=query)
    except AiContextApiError as ex:
        raise ToolError(str(ex))


__tools__ = (search_dlthub_sources,)



from typing import Any, Optional, Type, Union, Dict, TYPE_CHECKING

from dlt.common import logger
from dlt.common.data_types.typing import TDataType
from dlt.common.destination import Destination, DestinationCapabilitiesContext
from dlt.common.data_writers.escape import escape_redshift_identifier, escape_redshift_literal
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE

from dlt.common.destination.typing import PreparedTableSchema
from dlt.common.exceptions import TerminalValueError
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.schema.typing import TColumnSchema, TColumnType
from dlt.common.typing import TLoaderFileFormat

from dlt.destinations.type_mapping import TypeMapperImpl
from dlt.destinations.impl.redshift.configuration import (
    RedshiftCredentials,
    RedshiftClientConfiguration,
)

if TYPE_CHECKING:
    from dlt.destinations.impl.redshift.redshift import RedshiftClient
else:
    RedshiftClient = Any


class RedshiftTypeMapper(TypeMapperImpl):
    sct_to_unbound_dbt = {
        "json": "super",
        "text": "varchar(max)",
        "double": "double precision",
        "bool": "boolean",
        "date": "date",
        "timestamp": "timestamp with time zone",
        "bigint": "bigint",
        "binary": "varbinary",
        "time": "time without time zone",
    }

    sct_to_dbt = {
        "decimal": "numeric(%i,%i)",
        "wei": "numeric(%i,%i)",
        "text": "varchar(%i)",
        "binary": "varbinary(%i)",
    }

    dbt_to_sct = {
        "super": "json",
        "varchar(max)": "text",
        "double precision": "double",
        "boolean": "bool",
        "date": "date",
        "timestamp with time zone": "timestamp",
        "timestamp without time zone": "timestamp",
        "bigint": "bigint",
        "binary varying": "binary",
        "numeric": "decimal",
        "time without time zone": "time",
        "varchar": "text",
        "smallint": "bigint",
        "integer": "bigint",
    }

    def ensure_supported_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema,
        loader_file_format: TLoaderFileFormat,
    ) -> None:
        if loader_file_format in ("insert_values", "model"):
            return
        # time not supported on staging file formats
        if column["data_type"] == "time":
            raise TerminalValueError(
                "Please convert `datetime.time` objects in your data to `str` or"
                " `datetime.datetime`.",
                "time",
            )
        if loader_file_format == "jsonl":
            if column["data_type"] == "binary":
                raise TerminalValueError("", "binary")
        if loader_file_format == "parquet":
            # binary not supported on parquet if precision is set
            if column.get("precision") and column["data_type"] == "binary":
                raise TerminalValueError(
                    "Redshift cannot load fixed width VARBYTE columns from parquet files. Switch"
                    " to other file format or use binary columns without precision.",
                    "binary",
                )

    def to_db_datetime_type(
        self,
        column: TColumnSchema,
        table: PreparedTableSchema = None,
    ) -> str:
        column_name = column["name"]
        table_name = table["name"]
        timezone = column.get("timezone", True)
        precision = column.get("precision")

        if precision and precision > self.capabilities.max_timestamp_precision:
            logger.warn(
                f"Redshift only supports microsecond precision (6) for column '{column_name}' in"
                f" table '{table_name}'. Will use precision 6."
            )

        # Redshift has TIMESTAMPTZ for timezone-aware and TIMESTAMP for without timezone
        if timezone:
            return "timestamp with time zone"
        else:
            return "timestamp without time zone"

    def to_db_integer_type(self, column: TColumnSchema, table: PreparedTableSchema = None) -> str:
        precision = column.get("precision")
        if precision is None:
            return "bigint"
        if precision <= 16:
            return "smallint"
        elif precision <= 32:
            return "integer"
        elif precision <= 64:
            return "bigint"
        raise TerminalValueError(
            f"bigint with `{precision=:}` can't be mapped to Redshift integer type"
        )

    def from_destination_type(
        self, db_type: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        if db_type == "numeric":
            if (precision, scale) == self.capabilities.wei_precision:
                return dict(data_type="wei")
        if db_type == "timestamp without time zone":
            return {"data_type": "timestamp", "timezone": False}
        return super().from_destination_type(db_type, precision, scale)


class redshift(Destination[RedshiftClientConfiguration, "RedshiftClient"]):
    spec = RedshiftClientConfiguration

    def _raw_capabilities(self) -> DestinationCapabilitiesContext:
        caps = DestinationCapabilitiesContext()
        caps.preferred_loader_file_format = "insert_values"
        caps.supported_loader_file_formats = ["insert_values", "model"]
        caps.preferred_staging_file_format = "jsonl"
        caps.supported_staging_file_formats = ["jsonl", "parquet"]
        caps.type_mapper = RedshiftTypeMapper
        # redshift is case insensitive and will lower case identifiers when stored
        # you can enable case sensitivity https://docs.aws.amazon.com/redshift/latest/dg/r_enable_case_sensitive_identifier.html
        # then redshift behaves like postgres
        caps.escape_identifier = escape_redshift_identifier
        caps.escape_literal = escape_redshift_literal
        caps.casefold_identifier = str.lower
        caps.has_case_sensitive_identifiers = False
        caps.decimal_precision = (DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE)
        caps.wei_precision = (DEFAULT_NUMERIC_PRECISION, 0)
        caps.max_identifier_length = 127
        caps.max_column_identifier_length = 127
        caps.max_query_length = 16 * 1024 * 1024
        caps.is_max_query_length_in_bytes = True
        caps.max_text_data_type_length = 65535
        caps.is_max_text_data_type_length_in_bytes = True
        caps.supports_ddl_transactions = True
        caps.alter_add_multi_column = False
        caps.supported_merge_strategies = ["delete-insert", "scd2"]
        caps.supported_replace_strategies = ["truncate-and-insert", "insert-from-staging"]
        caps.sqlglot_dialect = "redshift"

        return caps

    @property
    def client_class(self) -> Type["RedshiftClient"]:
        from dlt.destinations.impl.redshift.redshift import RedshiftClient

        return RedshiftClient

    def __init__(
        self,
        credentials: Union[RedshiftCredentials, Dict[str, Any], str] = None,
        staging_iam_role: Optional[str] = None,
        has_case_sensitive_identifiers: bool = False,
        destination_name: str = None,
        environment: str = None,
        **kwargs: Any,
    ) -> None:
        """Configure the Redshift destination to use in a pipeline.

        All arguments provided here supersede other configuration sources such as environment variables and dlt config files.

        Args:
            credentials (Union[RedshiftCredentials, Dict[str, Any], str], optional): Credentials to connect to the redshift database. Can be an instance of `RedshiftCredentials` or
                a connection string in the format `redshift://user:password@host:port/database`. Defaults to None.
            staging_iam_role (Optional[str], optional): IAM role to use for staging data in S3. Defaults to None.
            has_case_sensitive_identifiers (bool, optional): Whether case sensitive identifiers are enabled for the database. Defaults to False.
            destination_name (str, optional): Name of the destination. Defaults to None.
            environment (str, optional): Environment name. Defaults to None.
            **kwargs (Any): Additional arguments passed to the destination config.
        """
        super().__init__(
            credentials=credentials,
            staging_iam_role=staging_iam_role,
            has_case_sensitive_identifiers=has_case_sensitive_identifiers,
            destination_name=destination_name,
            environment=environment,
            **kwargs,
        )

    @classmethod
    def adjust_capabilities(
        cls,
        caps: DestinationCapabilitiesContext,
        config: RedshiftClientConfiguration,
        naming: Optional[NamingConvention],
    ) -> DestinationCapabilitiesContext:
        # modify the caps if case sensitive identifiers are requested
        if config.has_case_sensitive_identifiers:
            caps.has_case_sensitive_identifiers = True
            caps.casefold_identifier = str
        return super().adjust_capabilities(caps, config, naming)


redshift.register()



from tenacity import RetryError
from requests import (
    Request,
    Response,
    ConnectionError,
    ConnectTimeout,
    FileModeWarning,
    HTTPError,
    ReadTimeout,
    RequestException,
    Timeout,
    TooManyRedirects,
    URLRequired,
)
from requests.exceptions import ChunkedEncodingError
from dlt.sources.helpers.requests.retry import Client
from dlt.sources.helpers.requests.session import Session

from dlt.common.configuration.inject import with_config
from dlt.common.configuration.specs import RuntimeConfiguration

# create initial instance without config injection
client = Client()
# wrap initializer to inject run configuration for custom clients
Client.__init__ = with_config(Client.__init__, spec=RuntimeConfiguration)  # type: ignore[method-assign]

get, post, put, patch, delete, options, head, request = (
    client.get,
    client.post,
    client.put,
    client.patch,
    client.delete,
    client.options,
    client.head,
    client.request,
)


def init(config: RuntimeConfiguration) -> None:
    """Initialize the default requests client from config"""
    client.update_from_config(config)


__all__ = [
    "client",
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "options",
    "head",
    "request",
    "init",
    "Session",
    "Request",
    "Response",
    "ConnectionError",
    "ConnectTimeout",
    "FileModeWarning",
    "HTTPError",
    "ReadTimeout",
    "RequestException",
    "Timeout",
    "TooManyRedirects",
    "URLRequired",
    "ChunkedEncodingError",
    "Client",
    "RetryError",
]



from dlt.destinations.impl.postgres.factory import postgres
from dlt.destinations.impl.snowflake.factory import snowflake
from dlt.destinations.impl.filesystem.factory import filesystem
from dlt.destinations.impl.duckdb.factory import duckdb
from dlt.destinations.impl.ducklake.factory import ducklake
from dlt.destinations.impl.dummy.factory import dummy
from dlt.destinations.impl.mssql.factory import mssql
from dlt.destinations.impl.fabric.factory import fabric
from dlt.destinations.impl.bigquery.factory import bigquery
from dlt.destinations.impl.athena.factory import athena
from dlt.destinations.impl.redshift.factory import redshift
from dlt.destinations.impl.qdrant.factory import qdrant
from dlt.destinations.impl.lancedb.factory import lancedb
from dlt.destinations.impl.motherduck.factory import motherduck
from dlt.destinations.impl.weaviate.factory import weaviate
from dlt.destinations.impl.destination.factory import destination
from dlt.destinations.impl.synapse.factory import synapse
from dlt.destinations.impl.databricks.factory import databricks
from dlt.destinations.impl.dremio.factory import dremio
from dlt.destinations.impl.clickhouse.factory import clickhouse
from dlt.destinations.impl.sqlalchemy.factory import sqlalchemy


__all__ = [
    "postgres",
    "snowflake",
    "filesystem",
    "duckdb",
    "ducklake",
    "dummy",
    "mssql",
    "fabric",
    "bigquery",
    "athena",
    "redshift",
    "qdrant",
    "lancedb",
    "motherduck",
    "weaviate",
    "synapse",
    "databricks",
    "dremio",
    "clickhouse",
    "destination",
    "sqlalchemy",
]
